"""

    Greynir: Natural language processing for Icelandic

    Dictionary-aware tokenization layer

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module adds layers on top of the "raw" tokenizer in
    tokenizer.py. These layers annotate the token stream with word
    meanings from the BIN lexicon of Icelandic, identify multi-word
    phrases, process person names, etc.

"""

from typing import (
    DefaultDict,
    Dict,
    cast,
    Optional,
    Tuple,
    List,
    Sequence,
    Mapping,
    Union,
    Iterable,
    Iterator,
    Set,
    FrozenSet,
    Callable,
    Type,
    Any,
    TypeVar,
)
from typing_extensions import TypedDict

import sys
import re
from collections import defaultdict

from tokenizer import (
    TOK,
    Tok,
    tokenize_without_annotation,
    # The following imports are here in order to be visible in clients
    # (they are not used in this module)
    correct_spaces,  # type: ignore
    paragraphs,  # type: ignore
    parse_tokens,  # type: ignore
)
from tokenizer.tokenizer import TokenStream
from tokenizer.definitions import (
    BIN_Tuple,
    BIN_TupleList,
    CurrencyTuple,
    NumberTuple,
    PersonNameList,
    PersonNameTuple,
    ValType,
    DateTimeTuple,
)
from tokenizer.abbrev import Abbreviations

from .settings import StaticPhrases, AmbigPhrases, DisallowedNames, NamePreferences
from .bindb import GreynirBin


class TokenDict(TypedDict, total=False):

    """ The type of a token dictionary returned from describe_token() """

    # Index in original token list
    ix: int
    # Kind
    k: int
    # Terminal
    t: str
    # Augmented terminal (optional)
    a: str
    # Meaning: ordmynd, ordfl, fl, beyging
    m: Tuple[str, str, str, str]
    # Text
    x: str
    # Value
    v: Any
    # Gender (for person tokens only)
    g: str
    # Error marker (optional)
    err: int


class CanonicalTokenDict(TypedDict, total=False):

    """ A token dictionary returned from canonicalize_token().
        This scheme is intended for external consumption,
        such as export in JSON format to clients. """

    # Index in original token list
    ix: int
    # Token kind, as a string (e.g. 'WORD')
    k: str
    # Terminal, normalized (e.g. 'no_kk_et_nf')
    t: str
    # Original terminal (e.g. '"og:st"')
    o: str
    # Augmented terminal (e.g. 'so_1_þf_gm_fh_nt')
    a: str
    # Text
    x: str
    # Lemma
    s: str
    # BÍN category ('kk', 'so', 'fs'...)
    c: str
    # BÍN fl field ('ism', 'ætt'...)
    f: str
    # BÍN inflection (beyging field, e.g. 'GM-FH-NT')
    b: str
    # Additional values, depending on token kind
    v: Union[str, float, Dict[str, Any]]


if "PyPy 7.3.0" in sys.version or "PyPy 7.2." in sys.version:
    # Patch bug in PyPy 7.2/7.3.0, which may raise an erroneous exception on str.rsplit()
    def all_except_suffix(s: str) -> str:  # type: ignore
        try:
            return s[0 : s.rindex(" ")]
        except ValueError:
            # String does not contain a space: return it whole
            return s


else:

    def all_except_suffix(s: str) -> str:
        return s.rsplit(maxsplit=1)[0]


# Generic type variable
T = TypeVar("T")
# The type of a list of tokens
TokenList = List[Tok]
TokenIterable = Iterable[Tok]
# The input argument type for the tokenize() function and derivatives thereof
StringIterable = Union[str, Iterable[str]]
# The type of a stream of tokens
TokenIterator = Iterator[Tok]
# The type of a tokenization pipeline phase
FirstPhaseFunction = Callable[[], TokenIterator]
FollowingPhaseFunction = Callable[[TokenIterator], TokenIterator]
PhaseFunction = Union[FirstPhaseFunction, FollowingPhaseFunction]
StateTuple = Tuple[List[str], int]
StateList = List[StateTuple]
StateDict = DefaultDict[str, StateList]
DisambiguationTuple = Tuple[str, FrozenSet[str]]
TokenConstructor = Type["Bin_TOK"]
FilterFunction = Callable[[BIN_Tuple], bool]

# Person names that are not recognized at the start of sentences
NOT_NAME_AT_SENTENCE_START: FrozenSet[str] = frozenset(
    (
        "Annar",
        "Annars",
        "Kalla",
        "Sanna",
        "Gamli",
        "Gamla",
        "Vinni",
        "Vinna",
        "Vilji",
        "Vilja",
        "Ljótur",
        "Ljót",
        "Ljóti",
        "Ljóts",
        "Mikill",
        "Mikil",
    )
)

# Set of all cases (nominative, accusative, dative, genitive)
ALL_CASES: FrozenSet[str] = frozenset(["nf", "þf", "þgf", "ef"])

# Genders
GENDER_SET: FrozenSet[str] = frozenset(("kk", "kvk", "hk"))
GENDER_DICT: Mapping[str, str] = {"KK": "kk", "KVK": "kvk", "HK": "hk"}

HYPHEN = "-"  # Normal hyphen
EN_DASH = "\u2013"  # "–"
EM_DASH = "\u2014"  # "—"
COMPOSITE_HYPHEN = EN_DASH
COMPOSITE_HYPHENS = HYPHEN + COMPOSITE_HYPHEN
HYPHEN_SPLIT_RE = r"[" + COMPOSITE_HYPHENS + r"]"

# Prefixes that can be applied to adjectives with an intervening hyphen
ADJECTIVE_PREFIXES: FrozenSet[str] = frozenset(["hálf", "marg", "semí", "full"])

# Recognize words that multiply numbers
MULTIPLIERS: Mapping[str, int] = {
    # "núll": 0,
    # "hálfur": 0.5,
    # "helmingur": 0.5,
    # "þriðjungur": 1.0 / 3,
    # "fjórðungur": 1.0 / 4,
    # "fimmtungur": 1.0 / 5,
    "einn": 1,
    "tveir": 2,
    "þrír": 3,
    "fjórir": 4,
    "fimm": 5,
    "sex": 6,
    "sjö": 7,
    "átta": 8,
    "níu": 9,
    "tíu": 10,
    "ellefu": 11,
    "tólf": 12,
    "þrettán": 13,
    "fjórtán": 14,
    "fimmtán": 15,
    "sextán": 16,
    "sautján": 17,
    "seytján": 17,
    "átján": 18,
    "nítján": 19,
    "tuttugu": 20,
    "þrjátíu": 30,
    "fjörutíu": 40,
    "fimmtíu": 50,
    "sextíu": 60,
    "sjötíu": 70,
    "áttatíu": 80,
    "níutíu": 90,
    # "par": 2,
    # "tugur": 10,
    # "tylft": 12,
    "hundrað": 100,
    "þúsund": 1000,  # !!! Bæði hk og kvk!
    "þús.": 1000,
    "milljón": 10 ** 6,
    "milla": 10 ** 6,
    "millj.": 10 ** 6,
    "milljarður": 10 ** 9,
    "miljarður": 10 ** 9,
    "ma.": 10 ** 9,
    "mrð.": 10 ** 9,
}

# The following must occur as lemmas in BÍN
DECLINABLE_MULTIPLIERS: FrozenSet[str] = frozenset(
    ("hundrað", "þúsund", "milljón", "milljarður")
)

# Recognize words for percentages
PERCENTAGES: Mapping[str, int] = {
    "prósent": 1,
    "prósenta": 1,
    "hundraðshluti": 1,
    "prósentustig": 1,
}

# Recognize words for nationalities (used for currencies)
NATIONALITIES: Mapping[str, str] = {
    "danskur": "dk",
    "enskur": "uk",
    "breskur": "uk",
    "bandarískur": "us",
    "kanadískur": "ca",
    "svissneskur": "ch",
    "sænskur": "se",
    "norskur": "no",
    "japanskur": "jp",
    "íslenskur": "is",
    "pólskur": "po",
    "kínverskur": "cn",
    "ástralskur": "au",
    "rússneskur": "ru",
    "indverskur": "in",
    "indónesískur": "id",
}

# Valid currency combinations
ISO_CURRENCIES: Mapping[Tuple[str, str], str] = {
    ("dk", "ISK"): "DKK",
    ("is", "ISK"): "ISK",
    ("no", "ISK"): "NOK",
    ("se", "ISK"): "SEK",
    ("uk", "GBP"): "GBP",
    ("us", "USD"): "USD",
    ("ca", "USD"): "CAD",
    ("au", "USD"): "AUD",
    ("ch", "CHF"): "CHF",
    ("jp", "JPY"): "JPY",
    ("po", "PLN"): "PLN",
    ("ru", "RUB"): "RUB",
    ("in", "INR"): "INR",  # Indian rupee
    ("id", "INR"): "IDR",  # Indonesian rupiah
    ("cn", "CNY"): "CNY",
    ("cn", "RMB"): "RMB",
}

# Amount abbreviations including 'kr' for the ISK
# Corresponding abbreviations are found in Abbrev.conf
AMOUNT_ABBREV: Mapping[str, int] = {
    "kr.": 1,
    "kr": 1,
    "þ.kr.": 10 ** 3,
    "þ.kr": 10 ** 3,
    "þús.kr.": 10 ** 3,
    "þús.kr": 10 ** 3,
    "m.kr.": 10 ** 6,
    "m.kr": 10 ** 6,
    "mkr.": 10 ** 6,
    "mkr": 10 ** 6,
    "ma.kr.": 10 ** 9,
    "ma.kr": 10 ** 9,
    "mrð.kr.": 10 ** 9,
    "mrð.kr": 10 ** 9,
}

# Number words can be marked as subjects (any gender) or as numbers
NUMBER_CATEGORIES: FrozenSet[str] = frozenset(["töl", "to", "kk", "kvk", "hk", "lo"])

# Recognize words for currencies
CURRENCIES: Mapping[str, str] = {
    "króna": "ISK",
    "ISK": "ISK",
    "kr.": "ISK",
    "kr": "ISK",
    "DKK": "DKK",
    "NOK": "NOK",
    "SEK": "SEK",
    "pund": "GBP",
    "sterlingspund": "GBP",
    "GBP": "GBP",
    "dollari": "USD",
    "dalur": "USD",
    "bandaríkjadalur": "USD",
    "USD": "USD",
    "franki": "CHF",
    "CHF": "CHF",
    "rúbla": "RUB",
    "RUB": "RUB",
    "rúpía": "INR",
    "INR": "INR",
    "IDR": "IDR",
    "jen": "JPY",
    "yen": "JPY",
    "JPY": "JPY",
    "zloty": "PLN",
    "PLN": "PLN",
    "júan": "CNY",
    "yuan": "CNY",
    "CNY": "CNY",
    "renminbi": "RMB",
    "RMB": "RMB",
    "evra": "EUR",
    "EUR": "EUR",
}

CURRENCY_GENDERS: Mapping[str, str] = {
    "ISK": "kvk",
    "DKK": "kvk",
    "NOK": "kvk",
    "SEK": "kvk",
    "GBP": "hk",
    "USD": "kk",
    "CHF": "kk",
    "RUB": "kvk",
    "INR": "kvk",
    "IDR": "kvk",
    "JPY": "hk",
    "PLN": "hk",
    "CNY": "hk",
    "RMB": "hk",
    "EUR": "kvk",
}

# Set of categories (fl fields in BÍN) that denote
# person names, Icelandic ('ism' or 'gæl') or foreign ('erm')
PERSON_NAME_SET: FrozenSet[str] = frozenset(("ism", "gæl", "erm"))

# Set of categories (fl fields in BÍN) for patronyms
# and matronyms, as well as gender-neutral family names
PATRONYM_SET: FrozenSet[str] = frozenset(("föð", "móð", "ætt"))

# Set of foreign middle names that start with a lower case letter
# ('Louis de Broglie', 'Jan van Eyck')
# 'of' was also here but caused problems
FOREIGN_MIDDLE_NAME_SET: FrozenSet[str] = frozenset(
    ("van", "de", "den", "der", "el", "al", "von", "la")
)

ENTITY_MIDDLE_NAME_SET: FrozenSet[str] = frozenset(
    ("in", "a", "an", "for", "and", "the", "for", "on", "of")
)

# Given names that can also be family names (and thus gender- and caseless as such)
BOTH_GIVEN_AND_FAMILY_NAMES: FrozenSet[str] = frozenset(("Hafstein",))

# Note: these must have a meaning for this to work, so specifying them
# as abbreviations in Abbrev.conf in Tokenizer is recommended
_CORPORATION_ENDINGS: FrozenSet[str] = frozenset(
    [
        "ehf.",
        "ehf",
        "hf.",
        "hses.",
        "hses",
        "hf",
        "bs.",
        "bs",
        "sf.",
        "sf",
        "slhf.",
        "slhf",
        "slf.",
        "slf",
        "svf.",
        "svf",
        "ohf.",
        "ohf",
        "Inc",
        "Inc.",
        "Incorporated",
        "Corp",
        "Corp.",
        "Corporation",
        "Ltd",
        "Ltd.",
        "Limited",
        "Co",
        "Co.",
        "Company",
        "Group",
        "AS",
        "ASA",
        "SA",
        "S.A.",
        "GmbH",
        "AG",
        "SARL",
        "S.à.r.l.",
    ]
)

# Abbreviations that we explicitly accept as a part of
# person names, if we are auto-capitalizing
MIDDLE_NAME_ABBREVS: FrozenSet[str] = frozenset(
    (
        "th",
        "kr",
        "st",
        "fr",
        "a",
        "á",
        "b",
        "c",
        "d",
        "e",
        "é",
        "f",
        "g",
        "h",
        "i",
        "í",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "ó",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "ú",
        "v",
        "w",
        "x",
        "y",
        "ý",
        "z",
        "þ",
        "æ",
        "ö",
    )
)

# Not name abbreviations if not followed by period or surname
NOT_NAME_ABBREVS: FrozenSet[str] = frozenset(("á", "í"))

# Words which should probably be lowercase
PREFER_LOWERCASE: FrozenSet[str] = frozenset(
    (
        "á",
        "bóndi",
        "frá",
        "ganga",
        "hæð",
        "mikil",
        "mikill",
    )
)


def load_token(*args: Any) -> Tuple[int, str, ValType]:
    """ Convert a plain, usually JSON serialized, argument tuple
        to kind, txt, val attributes """
    kind, txt, val = args[:3]
    if kind == TOK.WORD:
        val = [BIN_Tuple(*v) for v in val]
    elif kind == TOK.PERSON:
        val = [PersonNameTuple(*v) for v in val]
    else:
        val = tuple(val)
    return kind, txt, cast(ValType, val)


class Bin_TOK(TOK):

    """ Override the TOK class from tokenizer.py to allow a dummy
        token parameter to be passed into token constructors where
        required. This again allows errtokenizer.py in GreynirCorrect
        to add token error information. """

    @staticmethod
    def Word(
        t: Union[Tok, str],
        m: Optional[BIN_TupleList] = None,
        token: Union[None, Tok, Sequence[Tok]] = None,
    ) -> Tok:
        # Note that the m parameter cannot be easily type annotated,
        # as the Tokenizer package is still using a .pyi (Python 2.7-compatible)
        # type annotation scheme
        r = TOK.Word(t, m)
        if token is not None:
            # We are basing this word token on a previously existing
            # token or list of tokens: copy the original text field(s)
            if isinstance(token, Tok):
                r.original = token.original
            else:
                r.original = "".join(t.original or "" for t in token)
        return r

    @staticmethod
    def Number(
        t: Union[Tok, str],
        n: float,
        cases: Optional[List[str]] = None,
        genders: Optional[List[str]] = None,
        token: Optional[Tok] = None,
    ) -> Tok:
        return TOK.Number(t, n, cases, genders)

    @staticmethod
    def Amount(
        t: Union[Tok, str],
        iso: str,
        n: float,
        cases: Optional[List[str]] = None,
        genders: Optional[List[str]] = None,
        token: Optional[Tok] = None,
    ) -> Tok:
        return TOK.Amount(t, iso, n, cases, genders)

    @staticmethod
    def Person(
        t: Union[Tok, str],
        m: Optional[PersonNameList] = None,
        token: Optional[Tok] = None,
    ) -> Tok:
        return TOK.Person(t, m)


def annotate(
    db: GreynirBin,
    token_ctor: TokenConstructor,
    token_stream: TokenIterator,
    *,
    auto_uppercase: bool = False,
    no_sentence_start: bool = False
):
    """ Look up word forms in the BIN word database. If auto_uppercase
        is True, change lower case words to uppercase if it looks likely
        that they should be uppercase. If no_sentence_start is True,
        don't assume that the token stream starts a sentence. """

    at_sentence_start = False

    # Consume the iterable source in token_stream (which may be a generator)
    for t in token_stream:
        if t.kind != TOK.WORD:
            # Not a word: relay the token unchanged
            yield t
            if t.kind == TOK.S_BEGIN or t.punctuation == ":":
                # After an S_BEGIN, and also after a colon, we consider ourselves
                # to be at a sentence starting point - unless
                # no_sentence_start is set to True
                at_sentence_start = not no_sentence_start
            elif t.kind != TOK.PUNCTUATION and t.kind != TOK.ORDINAL:
                # Wait until we have something other than punctuation or an
                # ordinal number to conclude that the sentence has started
                at_sentence_start = False
            continue
        # This is a word token
        w = t.txt
        if not t.val:
            # Look up word in BÍN database
            # If word is found in PREFER_LOWERCASE we skip searching uppercase meanings
            # (if auto_uppercase is True)
            w, m = db.lookup_g(
                w, at_sentence_start, auto_uppercase and w not in PREFER_LOWERCASE
            )
            if not m:
                # No meaning found in BÍN
                # Check exceptional cases involving hyphens
                w = t.txt
                if w[0] in COMPOSITE_HYPHENS:
                    # Something like '-menn' in 'þingkonur og -menn'
                    _, m = db.lookup_g(w[1:], False, False)
                    if m:
                        m = [
                            BIN_Tuple(
                                # We leave the lemma intact here ('maður' for '-menn')
                                mm.stofn,
                                mm.utg,
                                mm.ordfl,
                                mm.fl,
                                # ...but keep the hyphen in the word form ('-menn')
                                w,
                                mm.beyging,
                            )
                            for mm in m
                        ]
                elif HYPHEN in w or COMPOSITE_HYPHEN in w:
                    # Word with embedded hyphen: 'marg-ítrekaðri',
                    # 'málfræði-greining'
                    parts = re.split(HYPHEN_SPLIT_RE, w)
                    # Start by checking whether it exists in BÍN without hyphens
                    w_new, m = "", []
                    if all(p[0].islower() for p in parts[1:]):
                        # ...but we only do this if all of the suffixes start
                        # with a lowercase character (so, we don't do this for
                        # 'Syðri-Hnaus' or 'Litla-Brekka')
                        w_new, m = db.lookup_g(
                            "".join(parts), at_sentence_start, auto_uppercase
                        )
                    if m:
                        # Found without hyphens: use that word form
                        m = [
                            BIN_Tuple(
                                # Leave the lemma intact (but it may already contain
                                # hyphens inserted by the compound word recognizer)
                                mm.stofn,
                                mm.utg,
                                mm.ordfl,
                                mm.fl,
                                # Keep the word form as it originally appeared,
                                # with its hyphens
                                w,
                                mm.beyging,
                            )
                            for mm in m
                        ]
                        # Emulate what auto_uppercase does in this case
                        if auto_uppercase and w_new[0].isupper():
                            w = w[0].upper() + w[1:]
                    else:
                        # Not found without hyphens:
                        # Look up the last part only
                        _, m = db.lookup_g(parts[-1], False, False)
                        if m:
                            m = [
                                BIN_Tuple(
                                    # In this case, keep the hyphens in the lemma,
                                    # imitating the compound word recognizer
                                    "-".join(parts[:-1] + [mm.stofn]),
                                    mm.utg,
                                    mm.ordfl,
                                    mm.fl,
                                    # Keep the word form intact with hyphens
                                    w,
                                    mm.beyging,
                                )
                                for mm in m
                            ]

            # Yield a word tuple with meanings
            yield token_ctor.Word(
                w if auto_uppercase else t.txt,
                m,
                token=t,
            )
        else:
            # Already have a meaning (most likely from an abbreviation that the
            # tokenizer has recognized)
            meanings = list(t.meanings)
            if not w.isupper() and " " not in w and "." not in w:
                # This token is not in all-caps and does not contain spaces or
                # periods. It is thus possible that it is an abbreviation that
                # could have additional meanings as a word in BÍN.
                w_new, m = db.lookup_g(w, at_sentence_start, auto_uppercase)
                if m:
                    # Additional meanings found: add them to
                    # the front of the meaning list, giving them a bit of
                    # priority over the dubious abbreviation
                    meanings = m + meanings
                    if auto_uppercase:
                        w = w_new
            yield token_ctor.Word(w, meanings, token=t)
        # We have yielded a word token: definitely no longer at sentence start
        at_sentence_start = False


def match_stem_list(
    token: Tok, stems: Mapping[str, T], filter_func: Optional[FilterFunction] = None,
) -> Optional[T]:
    """ Find the stem of a word token in given dict, or return None if not found """
    if token.kind != TOK.WORD:
        return None
    # Go through the meanings with their stems
    if token.val:
        for m in token.meanings:
            # If a filter function is given, pass candidates to it
            lower_stofn = m.stofn.lower()
            if lower_stofn in stems and (filter_func is None or filter_func(m)):
                return stems[lower_stofn]
    # No meanings found: this might be a foreign or unknown word
    # However, if it is still in the stems list we return True
    return stems.get(token.txt.lower(), None)


def case(bin_spec: str, default: str = "nf") -> str:
    """ Return the case specified in the bin_spec string """
    c = default
    if "NF" in bin_spec:
        c = "nf"
    elif "ÞF" in bin_spec:
        c = "þf"
    elif "ÞGF" in bin_spec:
        c = "þgf"
    elif "EF" in bin_spec:
        c = "ef"
    return c


def add_cases(cases: Set[str], bin_spec: str, default: str = "nf") -> None:
    """ Add the case specified in the bin_spec string, if any, to the cases set """
    c = case(bin_spec, default)
    if c:
        cases.add(c)


def all_cases(token: Tok, filter_func: Optional[FilterFunction] = None) -> List[str]:
    """ Return a list of all cases that the token can be in """
    cases: Set[str] = set()
    if token.kind == TOK.WORD and token.val:
        # Roll through the potential meanings and extract the cases therefrom
        for m in token.meanings:
            if filter_func is not None and not filter_func(m):
                continue
            if m.fl == "ob" or m.beyging == "-":
                # One of the meanings is an undeclined word: all cases apply
                cases = set(ALL_CASES)
                break
            add_cases(cases, m.beyging, "")
    return list(cases)


def all_common_cases(
    token1: Tok, token2: Tok, filter_func: Optional[FilterFunction] = None,
):
    """ Compute intersection of case sets for two tokens """
    set1 = set(all_cases(token1, filter_func))
    if not token2.val:
        # Token2 is not found in BÍN (probably an exotic currency name):
        # just return the cases of the first token
        return list(set1)
    set2 = set(all_cases(token2))
    return list(set1 & set2)


def all_genders(token: Tok) -> Optional[List[str]]:
    """ Return a list of the possible genders of the word in the token, if any """
    if token.kind != TOK.WORD:
        return None
    g: Set[str] = set()
    if token.val:

        def find_gender(m: BIN_Tuple) -> Optional[str]:
            if m.ordfl in GENDER_SET:
                return m.ordfl  # Plain noun
            # Probably number word ('töl' or 'to'): look at its spec
            for k, v in GENDER_DICT.items():
                if k in m.beyging:
                    return v
            return None

        for meaning in token.meanings:
            gn = find_gender(meaning)
            if gn is not None:
                g.add(gn)

    return list(g)


def parse_phrases_1(
    db: GreynirBin, token_ctor: TokenConstructor, token_stream: TokenIterator
) -> TokenIterator:
    """ Parse numbers and amounts """

    token: Tok = cast(Tok, None)

    try:

        # Maintain a one-token lookahead
        token = next(token_stream)
        while True:
            next_token: Tok = next(token_stream)

            # Logic for numbers that are partially or entirely
            # written out in words

            def number(tok: Tok):
                """ If the token denotes a number, return that number - or None """
                if tok.txt.lower() == "áttu":
                    # Do not accept 'áttu' (stem='átta', no kvk) as a number
                    return None
                return match_stem_list(
                    tok, MULTIPLIERS, filter_func=lambda m: m.ordfl in NUMBER_CATEGORIES
                )

            # Check whether we have an initial number word
            multiplier = number(token) if token.kind == TOK.WORD else None

            # Check for [number] 'hundred|thousand|million|billion'
            while (
                token.kind == TOK.NUMBER or multiplier is not None
            ) and next_token.kind == TOK.WORD:

                multiplier_next = number(next_token)

                def convert_to_num(token: Tok) -> Tok:
                    if multiplier is not None:
                        token = token_ctor.Number(
                            token.txt,
                            multiplier,
                            all_cases(token),
                            all_genders(token),
                            token=token,
                        )
                    return token

                if multiplier_next is not None:
                    # Retain the case of the last multiplier, except
                    # if it is genitive (eignarfall) and the previous
                    # token had a case ('hundruðum milljarða' is dative,
                    # not genitive)
                    next_case = all_cases(next_token)
                    next_gender = all_genders(next_token)
                    if "ef" in next_case:
                        # We may have something like 'hundruðum milljarða':
                        # use the case and gender of 'hundruðum', not 'milljarða'
                        next_case = all_cases(token) or next_case
                        next_gender = all_genders(token) or next_gender
                    token = convert_to_num(token)
                    # We send error information, if any, from the token
                    # into the freshly constructed composite token via
                    # the token parameter. This ensures that "fimmhundruð"
                    # is correctly marked with an error resulting from
                    # a split into "fimm" and "hundruð".
                    original = (token.original or "") + (next_token.original or "")
                    token = token_ctor.Number(
                        t=token.txt + " " + next_token.txt,
                        n=token.number * multiplier_next,
                        cases=next_case,
                        genders=next_gender,
                        token=token,
                    )
                    token.original = original
                    # Eat the multiplier token
                    next_token = next(token_stream)
                elif next_token.txt in AMOUNT_ABBREV:
                    # Abbreviations for ISK amounts
                    # For abbreviations, we do not know the case,
                    # but we try to retain the previous case information if any
                    token = convert_to_num(token)
                    num = cast(NumberTuple, token.val)
                    original = (token.original or "") + (next_token.original or "")
                    token = token_ctor.Amount(
                        token.txt + " " + next_token.txt,
                        "ISK",
                        num[0] * AMOUNT_ABBREV[next_token.txt],
                        num[1],
                        num[2],
                    )
                    token.original = original
                    next_token = next(token_stream)
                else:
                    # Check for [number] 'percent'
                    percentage = match_stem_list(next_token, PERCENTAGES)
                    if percentage is not None:
                        token = convert_to_num(token)
                        original = (token.original or "") + (next_token.original or "")
                        token = token_ctor.Percent(
                            token.txt + " " + next_token.txt,
                            token.number,
                            all_cases(next_token),
                            all_genders(next_token),
                        )
                        token.original = original
                        # Eat the percentage token
                        next_token = next(token_stream)
                    else:
                        break

                multiplier = None

            # Check for currency name doublets, for example
            # 'danish krona' or 'british pound'
            if token.kind == TOK.WORD and next_token.kind == TOK.WORD:
                nat = match_stem_list(token, NATIONALITIES)
                if nat is not None:
                    cur = match_stem_list(next_token, CURRENCIES)
                    if cur is not None:
                        if (nat, cur) in ISO_CURRENCIES:
                            # Match: accumulate the possible cases
                            iso_code = ISO_CURRENCIES[(nat, cur)]
                            # Filter the possible cases by considering adjectives
                            # having the correct form, i.e.
                            # strong inflection for indefinite nouns or
                            # weak inflection for definite nouns
                            if (
                                next_token.has_meanings
                                and "gr" in next_token.meanings[0].beyging
                            ):
                                # Definite form ('pundið', 'dollarinn')
                                form = "VB"
                            else:
                                # Indefinite form ('pund', 'dollari')
                                form = "SB"
                            original = (token.original or "") + (next_token.original or "")
                            token = token_ctor.Currency(
                                token.txt + " " + next_token.txt,
                                iso_code,
                                all_common_cases(
                                    token,
                                    next_token,
                                    lambda m: (m.ordfl == "lo" and form in m.beyging),
                                ),
                                [CURRENCY_GENDERS[cur]],
                            )
                            token.original = original
                            next_token = next(token_stream)

            # Check for composites:
            # 'stjórnskipunar- og eftirlitsnefnd'
            # 'dómsmála-, viðskipta- og iðnaðarráðherra'
            tq: List[Tok] = []
            while (
                token.kind == TOK.WORD or token.kind == TOK.ENTITY
            ) and next_token.punctuation == COMPOSITE_HYPHEN:
                tq.append(token)
                hyphen = TOK.Punctuation(next_token.txt, normalized=HYPHEN)
                hyphen.original = next_token.original
                tq.append(hyphen)
                # Check for optional comma after the prefix
                comma_token: Tok = next(token_stream)
                if comma_token.punctuation == ",":
                    # A comma is present: append it to the queue
                    # and skip to the next token
                    tq.append(comma_token)
                    comma_token = next(token_stream)
                # Reset our two lookahead tokens
                token = comma_token
                next_token = next(token_stream)

            if tq:
                # We have accumulated one or more prefixes
                # ('dómsmála-, viðskipta-')
                if token.kind == TOK.WORD and token.txt in ("og", "eða"):
                    # We have 'viðskipta- og'
                    if next_token.kind != TOK.WORD:
                        # Incorrect: yield the accumulated token
                        # queue and keep the current token and the
                        # next_token lookahead unchanged
                        for t in tq:
                            yield t
                    else:
                        # We have 'viðskipta- og iðnaðarráðherra'
                        # or 'dómsmála-, viðskipta- og iðnaðarráðherra'.
                        # Return a single token with the meanings of
                        # the last word, but an amalgamated token text.
                        # Note: there is no meaning check for the first
                        # part of the composition, so it can be an unknown word.
                        all_tq = tq + [token, next_token]
                        txt = " ".join(t.txt for t in all_tq)
                        txt = txt.replace(" -", "-").replace(" ,", ",")
                        # Create a fresh list of meanings with the full
                        # prefix in the ordmynd field
                        prefix = all_except_suffix(txt)
                        m = [
                            BIN_Tuple(
                                prefix + " " + mm.stofn,
                                mm.utg,
                                mm.ordfl,
                                mm.fl,
                                prefix + " " + mm.ordmynd,
                                mm.beyging,
                            )
                            for mm in cast(Tuple[BIN_Tuple, ...], next_token.val)
                        ]
                        # Copy attributes, such as capitalization status
                        # (cf. GreynirCorrect) from the first token in the queue,
                        # but send the whole queue so that the original string
                        # can be preserved
                        token = token_ctor.Word(txt, m, token=all_tq)
                        next_token = next(token_stream)
                else:
                    # Incorrect prediction: make amends and continue
                    for t in tq:
                        yield t

            # Yield the current token and advance to the lookahead
            yield token
            token = next_token

    except StopIteration:
        pass

    # Final token (previous lookahead)
    if token:
        yield token


def parse_phrases_2(
    token_stream: TokenIterator, token_ctor: TokenConstructor, auto_uppercase: bool
) -> TokenIterator:
    """ Parse a stream of tokens looking for phrases and making substitutions.
        Second pass: handle conversion of numbers + currencies into amounts,
        and process person names """

    token: Tok = cast(Tok, None)
    next_token: Tok = cast(Tok, None)

    try:
        # Use TokenStream wrapper for iterator with lookahead
        token_stream = TokenStream(token_stream)

        token = next(token_stream)
        # Maintain a set of full person names encountered
        names: Set[PersonNameTuple] = set()
        at_sentence_start = False

        while True:
            next_token = next(token_stream)

            # Make the lookahead checks we're interested in
            # Check for [number] [currency] and convert to [amount]
            if token.kind == TOK.NUMBER and (
                next_token.kind == TOK.WORD or next_token.kind == TOK.CURRENCY
            ):
                # Preserve the case of the number, if available
                # (milljónir, milljóna, milljónum)
                num = cast(NumberTuple, token.val)
                cases = num[1]
                genders = num[2]
                cur = None

                if next_token.kind == TOK.WORD:
                    # Try to find a currency name
                    cur = match_stem_list(next_token, CURRENCIES)
                    if cur is None and next_token.txt.isupper():
                        # Might be an ISO abbrev (which is not in BÍN)
                        cur = CURRENCIES.get(next_token.txt)
                        if not cases:
                            cases = list(ALL_CASES)
                        if not genders:
                            # Try to find a correct gender for the ISO abbrev,
                            # or use neutral as a default
                            genders = [CURRENCY_GENDERS.get(next_token.txt, "hk")]
                    if cur is not None:
                        # Use the case and gender information from the currency name
                        if not cases or "nf" in cases:
                            # No case information from the number, or
                            # it is nominative ('tólf hundruð pundum')
                            # use the case of the currency name only
                            new_cases = all_cases(next_token)
                            # However, for 'sex milljónum punda', we use
                            # the case of 'sex milljónum' and not 'punda'
                            if "ef" not in new_cases:
                                cases = new_cases
                        if not genders:
                            genders = all_genders(next_token)
                elif next_token.kind == TOK.CURRENCY:
                    # Already have an ISO identifier for a currency
                    ct = cast(CurrencyTuple, next_token.val)
                    cur = ct[0]
                    # Use the case and gender information from the currency name
                    # if no such information was given with the number itself
                    cases = cases or ct[1]
                    genders = genders or ct[2]

                if cur is not None:
                    # Create an amount
                    # Use the case and gender information from the number, if any
                    original = (token.original or "") + (next_token.original or "")
                    token = token_ctor.Amount(
                        token.txt + " " + next_token.txt, cur, num[0], cases, genders,
                        token=next_token,
                    )
                    token.original = original
                    # Eat the currency token
                    next_token = next(token_stream)

            # Check for [time] [date] (absolute)
            if token.kind == TOK.TIME and next_token.kind == TOK.DATEABS:
                # Create a time stamp
                h, m, s = cast(DateTimeTuple, token.val)
                y, mo, d = cast(DateTimeTuple, next_token.val)
                original = (token.original or "") + (next_token.original or "")
                token = token_ctor.Timestampabs(
                    token.txt + " " + next_token.txt, y=y, mo=mo, d=d, h=h, m=m, s=s
                )
                token.original = original
                # Eat the time token
                next_token = next(token_stream)

            # Check for [time] [date] (relative)
            if token.kind == TOK.TIME and next_token.kind == TOK.DATEREL:
                # Create a time stamp
                h, m, s = cast(DateTimeTuple, token.val)
                y, mo, d = cast(DateTimeTuple, next_token.val)
                original = (token.original or "") + (next_token.original or "")
                token = token_ctor.Timestamprel(
                    token.txt + " " + next_token.txt, y=y, mo=mo, d=d, h=h, m=m, s=s
                )
                token.original = original
                # Eat the time token
                next_token = next(token_stream)

            # Logic for human names

            def stems(
                tok: Tok, categories: FrozenSet[str], given_name: bool = False
            ) -> Optional[List[PersonNameTuple]]:
                """ If the token denotes a given name, return its possible
                    interpretations, as a list of PersonName tuples (name, case, gender).
                    If given_name is True, we omit from the list all name forms that
                    occur in the disallowed_names section in the configuration file. """
                if tok.kind != TOK.WORD or not tok.val:
                    return None
                if at_sentence_start and tok.txt in NOT_NAME_AT_SENTENCE_START:
                    # Disallow certain person names at the start of sentences,
                    # such as 'Annar'
                    return None
                # Set up the names we're not going to allow
                dstems = DisallowedNames.STEMS if given_name else {}
                # Look through the token meanings
                result: List[PersonNameTuple] = []
                for m in tok.meanings:
                    if m.fl in categories and ("ET" in m.beyging or m.beyging == "-"):
                        # If this is a given name, we cut out name forms
                        # that are frequently ambiguous and wrong,
                        # i.e. "Frá" as accusative of the name "Frár",
                        # and "Sigurð" in the nominative.
                        c = case(m.beyging, default="-")
                        if m.stofn not in dstems or c not in dstems[m.stofn]:
                            # Note the stem ('stofn') and the gender from
                            # the word type ('ordfl')
                            result.append(
                                PersonNameTuple(name=m.stofn, gender=m.ordfl, case=c)
                            )
                return result or None

            def has_category(tok: Tok, categories: FrozenSet[str]) -> bool:
                """ Return True if the token matches a meaning
                    with any of the given categories """
                if tok.kind != TOK.WORD or not tok.val:
                    return False
                return any(m.fl in categories for m in tok.meanings)

            def has_other_meaning(tok: Tok, categories: FrozenSet[str]) -> bool:
                """ Return True if the token can denote something
                    besides a given name """
                if tok.kind != TOK.WORD or not tok.val:
                    return True
                # Return True if there is a different meaning, not a given name
                return any(m.fl not in categories for m in tok.meanings)

            # Check for person names
            def given_names(tok: Tok) -> Optional[List[PersonNameTuple]]:
                """ Check for Icelandic or foreign person name
                    (category 'ism', 'gæl' or 'erm') """
                if tok.kind != TOK.WORD or not tok.txt[0].isupper():
                    # Must be a word starting with an uppercase character
                    return None
                return stems(tok, PERSON_NAME_SET, given_name=True)

            # Check for surnames
            def surnames(tok: Tok) -> Optional[List[PersonNameTuple]]:
                """ Check for Icelandic patronym (category 'föð'),
                    matronym (category 'móð') or family names (category 'ætt') """
                if tok.kind != TOK.WORD or not tok.txt[0].isupper():
                    # Must be a word starting with an uppercase character
                    return None
                return stems(tok, PATRONYM_SET)

            # Check for unknown surnames
            def unknown_surname(tok: Tok) -> bool:
                """ Check for unknown (non-Icelandic) surnames """
                # Accept (most) upper case words as a surnames
                if auto_uppercase:
                    if tok.txt and not tok.val:
                        # Looks like an unknown word: accept it as a surname
                        # (might be a foreign name)
                        return True
                if tok.kind != TOK.WORD or not tok.txt[0].isupper():
                    # Must start with capital letter
                    return False
                if has_category(tok, PATRONYM_SET):
                    # This is a known surname, not an unknown one
                    return False
                if tok.txt in _CORPORATION_ENDINGS:
                    return False
                # Allow single-letter abbreviations, but not multi-letter
                # all-caps words (those are probably acronyms)
                return len(tok.txt) == 1 or not tok.txt.isupper()

            def given_names_or_middle_abbrev(
                tok: Tok,
            ) -> Optional[List[PersonNameTuple]]:
                """ Check for given name or middle abbreviation """
                gnames = given_names(tok)
                wrd = tok.txt
                if gnames is not None:
                    if wrd in BOTH_GIVEN_AND_FAMILY_NAMES:
                        # For instance "Hafstein" which can be both a given
                        # name and a family name: prepend the family name as
                        # an genderless and caseless option to the list
                        gnames = [
                            PersonNameTuple(name=wrd, gender=None, case=None)
                        ] + gnames
                    return gnames

                if tok.kind != TOK.WORD:
                    return None

                if auto_uppercase and wrd.rstrip(".") in MIDDLE_NAME_ABBREVS:
                    # Capitalize middle name abbreviations
                    wrd = wrd.capitalize()
                    # Also update token txt
                    tok.txt = wrd

                # If wrd (without following period) is longer than
                # middle name abbrevs such as "th", "kr" or "f"
                # or not a foreign middle name (like "al", "der", "van")
                elif (
                    len(wrd.rstrip(".")) > 2 or wrd[0].islower()
                ) and wrd not in FOREIGN_MIDDLE_NAME_SET:
                    return None

                # Either:
                # - One or two letter middle name abbreviation (possibly with following period)
                # - Lowercase foreign middle name ("Thomas de Broglie", "Ruud van Nistelrooy")
                return [PersonNameTuple(name=wrd, gender=None, case=None)]

            def compatible(pn: PersonNameTuple, npn: PersonNameTuple) -> bool:
                """ Return True if the next PersonNameTuple (npn) is compatible
                    with the one we have (pn) """
                # The neutral gender (hk) is used for family names and is
                # compatible with both masculine and feminine given names
                if npn.gender and npn.gender != "hk" and (npn.gender != pn.gender):
                    return False
                if npn.case and npn.case != "-" and (npn.case != pn.case):
                    return False
                return True

            gn: Optional[List[PersonNameTuple]] = None
            # Accumulated original token text for person name
            namespan: str = ""
            if (
                token.kind == TOK.WORD
                and token.has_meanings
                and token.meanings[0].fl == "nafn"
            ):
                # Convert a WORD with fl="nafn" to a PERSON with the correct gender,
                # in all cases
                namespan = token.original or ""
                gender = token.meanings[0].ordfl
                token = token_ctor.Person(
                    token.txt,
                    m = [PersonNameTuple(token.txt, gender, case) for case in ALL_CASES],
                    token = token,
                )
                token.original = namespan
            else:
                gn = given_names(token)

            if gn:
                # Found at least one given name: look for a sequence of given names
                # having compatible genders and cases
                w: str = token.txt or ""
                namespan = token.original or ""
                patronym = False

                while True:
                    ntxt = next_token.txt

                    if auto_uppercase:
                        # Auto uppercasing of middle name abbreviations

                        # Interpret "s." as abbreviation of "símanúmer"
                        # when followed by telephone number
                        if ntxt == "s." and token_stream.kind() == TOK.TELNO:
                            # Followed by telephone number
                            # -> not a middle name abbreviation
                            break

                        # Other lowercase middle name abbreviations
                        if ntxt in MIDDLE_NAME_ABBREVS:

                            # Check if next token is a period
                            if token_stream.punctuation() == ".":
                                # Concatenate period token to middle name abbrev
                                # and remove period token
                                next_token = next_token.concatenate(next(token_stream))
                                ntxt = next_token.txt

                            elif (
                                ntxt in NOT_NAME_ABBREVS
                                and token_stream[0]
                                and not surnames(cast(Tok, token_stream[0]))
                            ):
                                # Next token is common word (such as "á", "í")
                                # and should only be considered a middle name
                                # if next word is a surname
                                break

                    # Deal with wrong sentence end/begin (S_END/S_BEGIN) tokens
                    # that sometimes appear in middle of sentence
                    # when middle name abbrevations end with period.
                    # However, if abbrev is last token in sentence, we keep the S_END token
                    if (
                        ntxt.endswith(".")
                        and ntxt.rstrip(".").lower() in MIDDLE_NAME_ABBREVS
                        and token_stream.kind(0) == TOK.S_END
                        and token_stream.kind(1) == TOK.S_BEGIN
                    ):
                        next(token_stream)  # Remove S_END
                        next(token_stream)  # Remove S_BEGIN

                    ngn = given_names_or_middle_abbrev(next_token)
                    if not ngn:
                        break
                    # Look through the stuff we got and see what is compatible
                    r: List[PersonNameTuple] = []
                    # pylint: disable=not-an-iterable
                    for p in gn:
                        # noinspection PyTypeChecker
                        for np in ngn:
                            if compatible(p, np):
                                # Compatible: add to result
                                r.append(
                                    PersonNameTuple(
                                        name=p.name + " " + np.name,
                                        gender=p.gender,
                                        case=p.case,
                                    )
                                )
                    if not r:
                        # This next name is not compatible with what we already
                        # have: break
                        break
                    # Success: switch to new given name list
                    gn = r
                    w += " " + next_token.txt
                    namespan += next_token.original or ""
                    next_token = next(token_stream)

                # Check whether the sequence of given names is followed
                # by one or more surnames (patronym/matronym) of the same gender,
                # for instance 'Dagur Bergþóruson Eggertsson'

                def eat_surnames(
                    gn: List[PersonNameTuple], w: str, patronym: bool, next_token: Tok
                ) -> Tuple[List[PersonNameTuple], str, bool, Tok]:
                    """ Process contiguous known surnames, typically "*dóttir/*son",
                        while they are compatible with the given name
                        we already have """
                    while True:
                        sn = surnames(next_token)
                        if not sn:
                            break
                        r: List[PersonNameTuple] = []
                        # Found surname: append it to the accumulated name,
                        # if compatible
                        for p in gn:
                            # pylint: disable=not-an-iterable
                            for np in sn:
                                if compatible(p, np):
                                    gender = (
                                        np.gender
                                        if (np.gender and np.gender != "hk")
                                        else p.gender
                                    )
                                    case = np.case if np.case != "-" else p.case
                                    r.append(
                                        PersonNameTuple(
                                            name=p.name + " " + np.name,
                                            gender=gender,
                                            case=case,
                                        )
                                    )
                        if not r:
                            break
                        # Compatible: include it and advance to the next token
                        gn = r
                        w += " " + next_token.txt
                        nonlocal namespan
                        namespan += next_token.original or ""
                        patronym = True
                        next_token = next(token_stream)
                    return gn, w, patronym, next_token

                gn, w, patronym, next_token = eat_surnames(gn, w, patronym, next_token)

                # Must have at least one possible name
                assert gn is not None
                assert len(gn) >= 1
                if not patronym:
                    # We stop name parsing after we find one or more Icelandic
                    # patronyms/matronyms. Otherwise, check whether we have an
                    # unknown uppercase word next;
                    # if so, add it to the person names we've already found
                    while unknown_surname(next_token):
                        ntxt = next_token.txt

                        if auto_uppercase and ntxt.islower():
                            ntxt = ntxt.capitalize()

                        for ix, p in enumerate(gn):
                            gn[ix] = PersonNameTuple(
                                name=p.name + " " + ntxt, gender=p.gender, case=p.case,
                            )
                        w += " " + ntxt
                        namespan += next_token.original or ""
                        next_token = next(token_stream)
                        # Assume we now have a patronym
                        patronym = True

                    if patronym:
                        # We still might have surnames coming up:
                        # eat them too, if present
                        gn, w, _, next_token = eat_surnames(gn, w, patronym, next_token)
                        assert gn is not None

                found_name = False
                # If we have a full name with patronym, store it
                if patronym:
                    names |= set(gn)
                else:
                    # Look through earlier full names and see whether this one matches
                    for ix, p in enumerate(gn):
                        gnames = p.name.split(" ")  # Given names
                        for lp in names:
                            match = (not p.gender) or (p.gender == lp.gender)
                            if match:
                                # The gender matches
                                # Leave the patronym off
                                lnames = set(lp.name.split(" ")[0:-1])
                                for n in gnames:
                                    if n not in lnames:
                                        # We have a given name that does not
                                        # match the person
                                        match = False
                                        break
                            if match:
                                # All given names match: assign the previously seen
                                # full name
                                gn[ix] = PersonNameTuple(
                                    name=lp.name, gender=lp.gender, case=p.case
                                )
                                found_name = True
                                break
                # If this is not a "strong" name, backtrack from recognizing it.
                # A "weak" name is (1) at the start of a sentence; (2) only one
                # word; (3) that word has a meaning that is not a name;
                # (4) the name has not been seen in a full form before;
                # (5) not on a 'well known name' list.
                weak = (
                    at_sentence_start
                    and (" " not in w)
                    and not patronym
                    and not found_name
                    and (
                        has_other_meaning(token, PERSON_NAME_SET)
                        and w not in NamePreferences.SET
                    )
                )

                if not weak:
                    # Return a person token with the accumulated name
                    # and the intersected set of possible cases
                    token = token_ctor.Person(w, gn)
                    token.original = namespan

            # Yield the current token and advance to the lookahead
            yield token
            if token.kind == TOK.S_BEGIN or token.punctuation == ":":
                at_sentence_start = True
            elif token.kind != TOK.PUNCTUATION and token.kind != TOK.ORDINAL:
                at_sentence_start = False
            token = next_token

    except StopIteration:
        pass

    # Final token (previous lookahead)
    if token:
        yield token


def parse_phrases_3(
    db: GreynirBin, token_stream: TokenIterator, token_ctor: TokenConstructor
) -> TokenIterator:
    """ Parse a stream of tokens looking for phrases and making substitutions.
        Third pass: coalesce uppercase, otherwise unrecognized words with
        a following person name, if any; also coalesce entity names and
        recognize company names by endings ('hf.', 'Inc.', etc.). """

    def is_interesting(token: Tok) -> bool:
        """ Return True if this token causes us to want to take
            a further look at the following tokens """
        if token.kind != TOK.ENTITY and token.kind != TOK.WORD:
            return False
        return token.txt[0].isupper()

    def can_concat(token: Tok) -> bool:
        """ Return True if the token content can be concatenated onto
            an existing entity name """
        # Non-capitalized function words that can appear within entity names
        if token.txt in ENTITY_MIDDLE_NAME_SET or token.txt in FOREIGN_MIDDLE_NAME_SET:
            return True
        if token.kind != TOK.ENTITY and token.kind != TOK.WORD:
            return False
        if not token.txt[0].isupper():
            return False
        if " " in token.txt:
            return False
        if token.kind == TOK.WORD and token.val:
            if any(m.stofn[0].isupper() for m in token.meanings):
                # This word has an independent uppercase meaning:
                # don't concatenate it
                return False
        return True

    def not_in_bin(token: Tok) -> bool:
        """ Return True if the token is not a normal word found in BÍN """
        if token.kind == TOK.ENTITY:
            return True
        assert token.kind == TOK.WORD
        if token.val:
            if all(m.ordfl != "entity" for m in token.meanings):
                # This word is found in BÍN and has no 'entity' meanings
                return False
        return True

    # The following declaration is a deliberate mypy hack
    token: Tok = cast(Tok, None)
    try:

        # Maintain a one-token lookahead
        token = next(token_stream)
        concatable = False

        while True:

            if not concatable and not is_interesting(token):
                if (
                    token.txt
                    and " " in token.txt
                    and token.txt.split(" ")[-1] in FOREIGN_MIDDLE_NAME_SET
                ):
                    # Combined in parse_phrases_2() but no capitalized word follows
                    # Should be split up
                    split = token.txt.split()
                    first = split[:-1]
                    middle = ""
                    orig = token.original
                    if first[-1] in FOREIGN_MIDDLE_NAME_SET:
                        # Allow one more check, in case of "de la"
                        middle = first[-1]
                        first = first[:-1]
                    if token.kind == TOK.PERSON:
                        token = token_ctor.Person(
                            " ".join(first),
                            [
                                PersonNameTuple(" ".join(first), pn.gender, pn.case)
                                for pn in token.person_names
                            ],
                        )
                    else:
                        token = token_ctor.Entity(" ".join(first))
                    token.original = orig
                    yield token
                    if middle:
                        _, m = db.lookup_g(middle)
                        x = token_ctor.Word(middle, m)
                        x.original = ""  # TODO: This could be made more intelligent
                        yield x
                    _, m = db.lookup_g(split[-1])
                    token = token_ctor.Word(split[-1], m)
                    token.original = ""  # TODO: This could be made more intelligent
                else:
                    yield token
                    # Make sure that token is None if next() raises StopIteration
                    token = cast(Tok, None)
                    token = next(token_stream)
                    continue
            next_token = next(token_stream)
            concatable = False

            if next_token.txt in _CORPORATION_ENDINGS:
                # Allow merging a corporation ending ('ehf.', 'Inc.'). This is fairly
                # open: any prefix consisting of uppercase words is
                # allowed, even if they are found in BÍN.
                original = (token.original or "") + (next_token.original or "")
                token = token_ctor.Company(token.txt + " " + next_token.txt)
                token.original = original
                next_token = next(token_stream)
            elif not_in_bin(token):
                if next_token.kind == TOK.PERSON and token.txt.istitle():
                    # Upper-case word that is either an entity or a word that is
                    # not in BÍN, and the next token is a person: merge the two
                    # tokens into a single person name
                    # 'Jesse' 'John Kelley' -> 'Jesse John Kelley'
                    original = (token.original or "") + (next_token.original or "")
                    token = token_ctor.Person(
                        token.txt + " " + next_token.txt,
                        [
                            PersonNameTuple(
                                token.txt + " " + pn.name, pn.gender, pn.case
                            )
                            for pn in next_token.person_names
                        ],
                    )
                    token.original = original
                    next_token = next(token_stream)
                elif can_concat(next_token):
                    # Concatenate the next token and do another loop round
                    original = (token.original or "") + (next_token.original or "")
                    token = token_ctor.Entity(token.txt + " " + next_token.txt)
                    token.original = original
                    concatable = True
                    continue

            # Yield the current token and advance to the lookahead
            yield token
            token = next_token

    except StopIteration:
        pass

    # Final token (previous lookahead)
    if token is not None:
        yield token


def fix_abbreviations(token_stream: TokenIterator) -> TokenIterator:
    """ Fix sentence splitting that may be wrong due to abbreviations """
    token: Tok = cast(Tok, None)
    try:
        # Maintain a one-token lookahead
        token = next(token_stream)
        while True:
            next_token = next(token_stream)
            # If we have a 'name finisher abbreviation'
            # (such as 'próf.' for 'prófessor') and the next token
            # is a text token but not a person, insert a sentence split
            if (
                token.kind == TOK.WORD
                and token.txt.lower() in Abbreviations.NAME_FINISHERS
                and next_token.kind in TOK.TEXT_EXCL_PERSON
            ):
                yield token
                yield TOK.End_Sentence()
                token = TOK.Begin_Sentence()
            # Yield the current token and advance to the lookahead
            yield token
            token = next_token
    except StopIteration:
        pass
    # Final token (previous lookahead)
    if token is not None:
        yield token


class MatchingStream:

    """ This class parses a stream of tokens while looking for
        multi-token matching sequences described in a phrase dictionary,
        and calling a matching function whenever those sequences
        occur in the stream, providing an opportunity to
        replace or modify these sequences.
    """

    def __init__(self, phrase_dictionary: StateDict) -> None:
        self._pdict = phrase_dictionary

    def key(self, token: Tok) -> Any:
        """ Generate a state key from the given token """
        return token.txt.lower()

    def match_state(self, key: Any, state: StateDict) -> StateList:
        """ Returns an iterable of states that match the key,
            or a falsy value if the key matches no states. """
        return state.get(key, [])

    def match(self, tq: List[Tok], ix: int) -> Iterable[Tok]:
        """ Called when we have found a match for the entire
            token queue tq, using the index ix """
        return tq

    def length(self, ix: int) -> int:
        """ Override this to provide the length of the actual
            phrase that matches at index ix """
        return 0

    def process(self, token_stream: TokenIterator) -> TokenIterator:
        """ Generate an output stream from the input token stream """
        # Token queue
        tq: List[Tok] = []
        # Phrases we're considering
        state: StateDict = defaultdict(list)
        pdict = self._pdict  # The phrase dictionary
        token: Optional[Tok]

        try:

            while True:

                token = next(token_stream)

                if token.txt is None:
                    # Not a word: no match; yield the token queue
                    if tq:
                        yield from tq
                        tq = []
                    # Discard the previous state, if any
                    if state:
                        state = defaultdict(list)
                    # ...and yield the non-matching token
                    yield token
                    continue

                # Look for matches in the current state and build a new state
                newstate: StateDict = defaultdict(list)
                key = self.key(token)

                def add_to_state(slist: List[str], index: int) -> None:
                    """ Add the list of subsequent words to the new parser state """
                    next_key = slist[0]
                    rest = slist[1:]
                    newstate[next_key].append((rest, index))

                def accept(state: List[Tuple[List[str], int]]) -> TokenIterator:
                    """ The current token matches the given state, either as
                        a continuation of a previous state or as an initiation
                        of a new phrase """
                    nonlocal token, newstate, tq
                    if token:
                        tq.append(token)
                        token = cast(Tok, None)
                    # sl is the continuation list (possible next tokens)
                    # for each state
                    for sl, ix in state:
                        if not sl:
                            # No continuation token from this state:
                            # this is a complete match
                            phrase_length = self.length(ix)
                            while len(tq) > phrase_length:
                                # We have extra queued tokens in the token queue
                                # that belong to a previously seen partial phrase
                                # that was not completed: yield them first
                                yield tq.pop(0)
                            if tq:
                                # Let the match function decide what to yield
                                # from this matched state
                                yield from self.match(tq, ix)
                                tq = []
                            # Make sure that we start from a fresh state and
                            # a fresh token queue when processing the next token
                            if newstate:
                                newstate = defaultdict(list)
                            # Note that it is possible to match even longer phrases
                            # by including a starting phrase in its entirety in
                            # the static phrase dictionary
                            break
                        # Nonempty continuation: add it to the next state
                        add_to_state(sl, ix)

                siter = self.match_state(key, state)
                if siter:
                    # This matches an expected token:
                    # go through potential continuations
                    yield from accept(siter)
                else:
                    # This matches no expected token, i.e. is not a
                    # continuation of any previously pushed state
                    if tq:
                        # Yield the accumulated token queue
                        yield from tq
                        tq = []
                    # Check whether this token starts a new phrase.
                    # Note: we cannot allow the last token of a
                    # previous phrase to start a new phrase, since it
                    # has already been consumed and acted upon
                    # (and, indeed, in that case the token variable
                    # would contain None at this point)
                    siter = self.match_state(key, pdict)
                    if siter:
                        # This word potentially starts a new phrase
                        yield from accept(siter)
                    elif token:
                        # Not starting a new phrase: pass the token through
                        yield token

                # Transition to the new state
                state = newstate

        except StopIteration:
            # Token stream is exhausted
            pass

        # Yield any tokens remaining in queue
        yield from tq


class StaticPhraseStream(MatchingStream):

    """ Process a stream of tokens looking for static multiword phrases
        (i.e. phrases that are not affected by inflection).
        The algorithm implements N-token lookahead where N is the
        length of the longest phrase.
    """

    def __init__(self, token_ctor: TokenConstructor, auto_uppercase: bool) -> None:
        super().__init__(StaticPhrases.DICT)
        self._token_ctor = token_ctor
        self._auto_uppercase = auto_uppercase

    def length(self, ix: int) -> int:
        return StaticPhrases.get_length(ix)

    def key(self, token: Tok) -> Tuple[str, str]:
        """ We allow both the original token text and a lowercase
            version of it to match """
        wo = token.txt  # Original word
        w = wo.lower()  # Lower case
        if w is not wo and (w == wo):
            wo = w
        return wo, w

    def match_state(self, key: Tuple[str, str], state: StateDict) -> StateList:
        """ First check for original (uppercase) word in the state, if any;
            if that doesn't match, check the lower case """
        wm = ""
        wo, w = key
        if self._auto_uppercase and len(wo) == 1 and w != wo:
            # If we are auto-uppercasing, leave single-letter lowercase
            # phrases alone, i.e. 'g' for 'gram' and 'm' for 'meter'
            wm = wo
        elif wo is not w and wo in state:
            wm = wo  # Original word
        elif w in state:
            wm = w  # Lowercase version
        return state.get(wm, [])

    def match(self, tq: List[Tok], ix: int) -> Iterable[Tok]:
        w = " ".join([t.txt for t in tq])
        # Add the entire phrase as one 'word' to the token queue.
        # Note that the StaticPhrases meaning list will be converted
        # to BIN_Tuple tuples in the annotate() pass.
        # Also note that the entire token queue is sent in as
        # the token parameter, as any token in the queue may
        # contain error information.
        newtok = self._token_ctor.Word(w, StaticPhrases.get_meaning(ix), token=tq)
        yield newtok


def parse_static_phrases(
    token_stream: TokenIterator, token_ctor: TokenConstructor, auto_uppercase: bool
) -> TokenIterator:
    """ Use the StaticPhraseStream class to process the token stream
        and replace static phrases with single tokens """
    sps = StaticPhraseStream(token_ctor, auto_uppercase)
    return sps.process(token_stream)


class DisambiguationStream(MatchingStream):

    """ Disambiguates a token stream by only retaining word
        meanings that have categories matching those allowed
        in the [disambiguate_phrases] section in config/Phrases.conf """

    def __init__(self, token_ctor: TokenConstructor) -> None:
        super().__init__(AmbigPhrases.DICT)
        self._token_ctor = token_ctor

    def key(self, token: Tok) -> DisambiguationTuple:
        """ Generate a phrase key from the given token """
        # Construct a set of all possible lemmas of this word form
        if token.kind == TOK.WORD:
            return token.txt.lower(), frozenset(m.stofn + "*" for m in token.meanings)
        return token.txt.lower(), frozenset()

    def match_state(self, key: DisambiguationTuple, state: StateDict) -> StateList:
        """ Called to see if the current token's key matches
            the given state. Returns the value that should be
            used to look up the key within the state, or None
            if there is no match. """
        # First, check for a direct text match
        txt, stems = key
        states = list(state.get(txt, []))
        # Then, check whether the stems of the token match any
        # asterisk-marked entry in the state
        for stem in stems:
            if stem in state:
                states.extend(state[stem])
        return states

    def length(self, ix: int) -> int:
        return len(AmbigPhrases.get_cats(ix))

    def match(self, tq: List[Tok], ix: int) -> Iterable[Tok]:
        """ We have a phrase match: return the tokens in the token
            queue, but with their meanings filtered down to only
            the word categories specified in the phrase configration """
        cats = AmbigPhrases.get_cats(ix)
        words = AmbigPhrases.get_words(ix)
        token_ctor = self._token_ctor
        assert len(tq) == len(cats)
        for t, cat_set, word in zip(tq, cats, words):
            # Yield a new token with fewer meanings for each
            # original token in the queue
            if t.kind != TOK.WORD or "*" in cat_set:
                # Not a word or no category constraint:
                # nothing to do
                yield t
                continue
            # Prepare the constrained list of meanings
            if "fs" in cat_set:
                # Handle prepositions specially, since we may have additional
                # preps defined in Main.conf that don't have fs meanings in BÍN
                w = t.txt.lower()
                mm = [BIN_Tuple(w, 0, "fs", "alm", w, "-")]
                cat_set = cat_set - frozenset(("fs",))
                # !!! BUG: constraining the meanings of prepositions (ordfl=fs)
                # !!! isn't currently meaningful, since the matcher in binparser.py
                # !!! for fs terminals doesn't look at the token meanings
            else:
                mm = []
            if cat_set:
                # Eliminate meanings that are not in the allowed category
                # set, or that have stems that don't match a stem specification
                # (i.e. phrase components marked with an asterisk, such as 'eiga*')
                stem = word[:-1] if word[-1] == "*" else None
                mm.extend(
                    m
                    for m in t.meanings
                    if m.ordfl in cat_set and (stem is None or m.stofn == stem)
                )

            yield token_ctor.Word(t.txt, mm, token=t)


def disambiguate_phrases(
    token_stream: TokenIterator, token_ctor: TokenConstructor
) -> TokenIterator:

    """ Parse a stream of tokens looking for common ambiguous multiword phrases
        (i.e. phrases that have a well known very likely interpretation but
        other extremely uncommon ones are also grammatically correct).
    """

    ds = DisambiguationStream(token_ctor)
    yield from ds.process(token_stream)


class DefaultPipeline:

    """ A DefaultPipeline encapsulates a sequence of tokenization
        phases, with each phase being a generator that accepts
        tokens from its input stream and yields tokens on its
        output stream. Individual phases in the sequence can
        easily be overridden in derived classes. """

    _token_ctor: TokenConstructor = Bin_TOK

    def __init__(self, text_or_gen: StringIterable, **options: Any) -> None:
        self._text_or_gen = text_or_gen
        self._auto_uppercase: bool = options.pop("auto_uppercase", False)
        self._no_sentence_start: bool = options.pop("no_sentence_start", False)
        self._options = options
        self._db: Optional[GreynirBin] = None
        # Initialize the default tokenizer pipeline.
        # This sequence of phases can be modified in derived classes.
        self._phases: List[PhaseFunction] = [
            self.tokenize_without_annotation,
            self.correct_tokens,
            self.parse_static_phrases,
            self.annotate,
            self.recognize_entities,
            self.check_spelling,
            self.parse_phrases_1,
            self.parse_phrases_2,
            self.parse_phrases_3,
            self.fix_abbreviations,
            self.disambiguate_phrases,
            self.final_correct,
        ]

    def tokenize_without_annotation(self) -> TokenIterator:
        """ The basic, raw tokenization from the tokenizer package """
        return tokenize_without_annotation(self._text_or_gen, **self._options)

    def parse_static_phrases(self, stream: TokenIterator) -> TokenIterator:
        """ Static multiword phrases """
        return parse_static_phrases(stream, self._token_ctor, self._auto_uppercase)

    def correct_tokens(self, stream: TokenIterator) -> TokenIterator:
        """ Token-level correction can be plugged in here (default stack doesn't do
            any corrections, but this is overridden in GreynirCorrect) """
        return stream

    def annotate(self, stream: TokenIterator) -> TokenIterator:
        """ Lookup meanings from dictionary """
        assert self._db is not None
        return annotate(
            self._db,
            self._token_ctor,
            stream,
            auto_uppercase=self._auto_uppercase,
            no_sentence_start=self._no_sentence_start,
        )

    def recognize_entities(self, stream: TokenIterator) -> TokenIterator:
        """ Recognize named entities. Default stack doesn't do anything,
            but derived classes can override this. """
        return stream

    def check_spelling(self, stream: TokenIterator) -> TokenIterator:
        """ Spelling correction can be plugged in here (default stack doesn't do
            any corrections, but this is overridden in GreynirCorrect) """
        return stream

    def parse_phrases_1(self, stream: TokenIterator) -> TokenIterator:
        """ Numbers and amounts """
        assert self._db is not None
        return parse_phrases_1(self._db, self._token_ctor, stream)

    def parse_phrases_2(self, stream: TokenIterator) -> TokenIterator:
        """ Currencies, person names """
        return parse_phrases_2(stream, self._token_ctor, self._auto_uppercase)

    def parse_phrases_3(self, stream: TokenIterator) -> TokenIterator:
        """ Additional person and entity name logic """
        assert self._db is not None
        return parse_phrases_3(self._db, stream, self._token_ctor)

    def fix_abbreviations(self, stream: TokenIterator) -> TokenIterator:
        """ Fix sentence splitting relating to abbreviations """
        return fix_abbreviations(stream)

    def disambiguate_phrases(self, stream: TokenIterator) -> TokenIterator:
        """ Eliminate very uncommon meanings """
        return disambiguate_phrases(stream, self._token_ctor)

    def final_correct(self, stream: TokenIterator) -> TokenIterator:
        """ Late-stage token correction, overridden in GreynirCorrect """
        return stream

    def tokenize(self) -> TokenIterator:
        """ Tokenize text in several phases, returning a generator of tokens
            that processes the text on-demand. If auto_uppercase is True, the tokenizer
            attempts to correct lowercase words that probably should be uppercase.
            If correction_func is not None, the given generator function is inserted
            into the token chain after processing static phrases but before
            BÍN annotation. This gives an opportunity to perform context-independent
            spelling corrections and the like. """

        if self._db is not None:
            # This should never occur, even in multi-threaded programs, since
            # each tokenize() call creates its own instance of DefaultPipeline
            raise ValueError("DefaultPipeline.tokenize() is not re-entrant")

        # We stack the tokenization phases together. Each generator
        # becomes a consumer of the previous generator.

        # Thank you Python for enabling this programming pattern ;-)

        with GreynirBin.get_db() as db:
            try:
                self._db = db
                # First tokenization phase
                token_stream = cast(FirstPhaseFunction, self._phases[0])()
                # Stack the other phases on top of each other
                for phase in self._phases[1:]:
                    token_stream = cast(FollowingPhaseFunction, phase)(token_stream)
                # ...and return the resulting chained generator
                return token_stream
            finally:
                self._db = None


def tokenize(text: StringIterable, **options: Any) -> TokenIterator:
    """ Tokenize text using the default pipeline """
    pipeline = DefaultPipeline(text, **options)
    return pipeline.tokenize()


def tokens_are_foreign(tokens: TokenIterable, min_icelandic_ratio: float) -> bool:
    """ Return True if the given tokens are probably not in Icelandic """
    words_in_bin = 0
    words_not_in_bin = 0
    # Enumerate through the tokens
    for t in tokens:
        if t.kind == TOK.WORD:
            if t.val:
                # The word has at least one meaning
                words_in_bin += 1
            else:
                # The word has no recognized meaning
                words_not_in_bin += 1
        elif t.kind == TOK.PERSON:
            # Person names count as recognized words
            words_in_bin += 1
        elif t.kind == TOK.ENTITY:
            # Entity names do not count as recognized words;
            # we count each enclosed word in the entity name
            words_not_in_bin += t.txt.count(" ") + 1
    # Return True if the sentence has at least three words
    # but less than 60% of them are found in BÍN
    num_words = words_in_bin + words_not_in_bin
    return num_words > 2 and words_in_bin / num_words < min_icelandic_ratio


def stems_of_token(t: "TokenDict") -> List[Tuple[str, str]]:
    """ Return a list of word stem descriptors associated with the token t.
        This is an empty list if the token is not a word or person or entity name.
    """
    kind = t.get("k", TOK.WORD)
    if kind not in {TOK.WORD, TOK.PERSON, TOK.ENTITY}:
        # No associated stem
        return []
    if kind == TOK.WORD:
        # Obtain the stem and the word category from the 'm' (meaning) field,
        # if present
        m = t.get("m")
        if m:
            stem, cat = m[0], m[1]
            return [(stem, cat)]
        # Entity or unknown word: fall through
    elif kind == TOK.PERSON:
        # The full person name, in nominative case, is stored in the 'v' field
        stem = t["v"]
        if "t" in t:
            # The gender is at the end of the corresponding terminal name
            gender = "_" + t["t"].split("_")[-1]
        elif "g" in t:
            # No terminal: there might be a dedicated gender ('g') field
            gender = "_" + t["g"]
        else:
            # No known gender
            gender = ""
        return [(stem, "person" + gender)]
    # TOK.ENTITY or unknown word
    stem = t["x"]
    return [(stem, "entity")]


def choose_full_name(
    val: PersonNameList, case: Optional[str], gender: Optional[str]
) -> Tuple[str, str]:
    """ From a list of name possibilities in val, and given a case and a gender
        (which may be None), return the best matching full name and gender """
    fn_list = [
        (fn, g, c)
        for fn, g, c in val
        if (gender is None or g == gender) and (case is None or c == case)
    ]
    if not fn_list:
        # Oops - nothing matched this. Might be a foreign, undeclinable name.
        # Try nominative if it wasn't alredy tried
        if case is not None and case != "nf":
            fn_list = [
                (fn, g, c)
                for fn, g, c in val
                if (gender is None or g == gender) and (case == "nf")
            ]
        # If still nothing, try anything with the same gender
        if not fn_list and gender is not None:
            fn_list = [(fn, g, c) for fn, g, c in val if (g == gender)]
        # If still nothing, give up and select the first available meaning
        if not fn_list:
            fn, g, c = val[0]
            fn_list = [(fn, g, c)]
    # If there are many choices, select the nominative case,
    # or the first element as a last resort
    ft = next((fn for fn in fn_list if fn[2] == "nf"), fn_list[0])
    return ft[0], (ft[1] if gender is None else gender) or "hk"



"""

    Greynir: Natural language processing for Icelandic

    Grammar module

    Copyright (C) 2021 Miðeind ehf.

    This software is licensed under the MIT License:

        Permission is hereby granted, free of charge, to any person
        obtaining a copy of this software and associated documentation
        files (the "Software"), to deal in the Software without restriction,
        including without limitation the rights to use, copy, modify, merge,
        publish, distribute, sublicense, and/or sell copies of the Software,
        and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
        EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
        MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
        IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
        CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
        TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
        SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    This module parses a text file containing the description of
    a context-free grammar.

    A grammar is specified as a set of rules. Each rule has a single
    left-hand-side nonterminal, associated with 1..n right-hand-side
    productions. Each right-hand-side production is a sequence of
    nonterminals and terminals. A terminal can match a token of
    input.

    In Greynir grammars, nonterminals always start with an uppercase letter.
    Terminals may be identifiers starting with lowercase letters, or
    literals enclosed within single or double quotes. Epsilon (empty)
    productions are allowed and denoted by 0.

    Examples:

    # Sequence (where terminal_X is optional at the start)
    Nonterminal_A -> terminal_X? Nonterminal_B terminal_Y

    # Alternatives (where terminal_W may occur one or more times sequentially)
    Nonterminal_B ->
        terminal_Z | terminal_W+ | Nonterminal_C

    # Literal terminals:
    # 'x' matches words with stem 'x'
    # "y" matches the token 'y' exactly (after conversion to lower case)
    Nonterminal_C ->
        Nonterminal_D | 'literal_1'* | "exact_literal_2"+

    # Epsilon
    Nonterminal_D ->
        'literal_2' | 0   # 0 means epsilon (in this case equivalent to 'literal_2'?)

"""

from typing import (
    Callable,
    DefaultDict,
    FrozenSet,
    List,
    Dict,
    Set,
    TYPE_CHECKING,
    Tuple,
    Iterable,
    Iterator,
    Optional,
    Union,
    Any,
    cast,
)

import os
import struct

from datetime import datetime
from collections import defaultdict, OrderedDict

# pylint: disable=no-name-in-module
if TYPE_CHECKING or __package__:
    from .settings import Settings, StaticPhrases
    from .basics import changedlocale, ConfigError
else:
    from settings import Settings, StaticPhrases
    from basics import changedlocale, ConfigError


ProductionTuple = Tuple[int, "Production"]


class GrammarError(Exception):

    """ Exception class for errors in a grammar """

    def __init__(self, text: str, fname: Optional[str] = None, line: int = 0) -> None:
        """ A GrammarError contains an error text and optionally the name
            of a grammar file and a line number where the error occurred """
        super().__init__(text)
        self.fname = fname
        self.line = line

    def augment(self, fname: str, line: int) -> None:
        """ Add filename and line information, if missing """
        if self.fname is None:
            self.fname = fname
        if self.line == 0:
            self.line = line

    def __str__(self) -> str:
        """ Create a string representation showing the file name and
            line number where the error originated, if available """
        prefix = ""
        if self.line:
            prefix = "Line " + str(self.line) + ": "
        if self.fname:
            prefix = self.fname + " - " + prefix
        return prefix + super().__str__()


class GrammarItem:

    """ An item that can occur in a production in a grammar.
        This is either a Nonterminal or a Terminal. """

    def __init__(self) -> None:
        self._index = 0

    @property
    def literal_text(self) -> str:
        """ The literal text of a terminal, if it is a literal terminal """
        return ""

    @property
    def index(self) -> int:
        """ Return the sequence number of this grammar item """
        return self._index

    def set_index(self, ix: int) -> None:
        """ Set a new sequence number for this grammar item """
        self._index = ix


class Nonterminal(GrammarItem):

    """ A nonterminal, either at the left hand side of
        a rule or within a production """

    _index = -1  # Running sequence number (negative) of all nonterminals

    # If no_reduce is True for a nonterminal, its child families are
    # not reduced to the single highest-scoring family after parsing.
    # This feature is used in query processing.
    no_reduce: bool = False

    def __init__(self, name: str, fname: Optional[str] = None, line: int = 0) -> None:
        super().__init__()
        self._name = name
        # Place of initial definition in a grammar file
        self._fname = fname
        self._line = line
        # Tags for this nonterminal
        self._tags: Optional[Set[str]] = None
        # Has this nonterminal been referenced in a production?
        self._ref = False
        # Is this an optional nonterminal, i.e. one that is
        # explicitly nullable?
        self._optional = name.endswith("?") or name.endswith("*")
        # Give all nonterminals a unique, negative sequence number for hashing purposes
        self._index = Nonterminal._index
        Nonterminal._index -= 1
        self._hash = id(self).__hash__()

    def __hash__(self) -> int:
        """ Use the id of this nonterminal as a basis for the hash """
        # The index may change after the entire grammar has been
        # read and processed; therefore it is not suitable for hashing
        return self._hash

    def __eq__(self, o: Any) -> bool:
        return self is o

    def __ne__(self, o: Any) -> bool:
        return self is not o

    def set_index(self, ix: int) -> None:
        """ Set a new sequence number for this nonterminal """
        # Nonterminals have negative indices
        assert ix < 0
        super().set_index(ix)

    def add_ref(self) -> None:
        """ Mark this as being referenced """
        self._ref = True

    @property
    def has_ref(self) -> bool:
        """ Return True if the nonterminal has been referenced in a production """
        return self._ref

    @property
    def is_optional(self) -> bool:
        """ Return True if this nonterminal is explicitly nullable """
        return self._optional

    @property
    def name(self) -> str:
        return self._name

    @property
    def fname(self) -> str:
        """ Return the name of the grammar file where this nonterminal was defined """
        return self._fname or ""

    @property
    def line(self) -> int:
        """ Return the number of the line within the grammar file
            where this nt was defined """
        return self._line

    @property
    def has_tags(self) -> bool:
        """ Return True if this nonterminal has a tag associated with it """
        return bool(self._tags)

    def has_tag(self, tag: str) -> bool:
        """ Check whether this nonterminal has been tagged with the given tag """
        return self._tags is not None and tag in self._tags

    def has_any_tag(self, tagset: FrozenSet[str]) -> bool:
        """ Check whether this nonterminal has been tagged
            with any of the given tags """
        return False if self._tags is None else not self._tags.isdisjoint(tagset)

    def add_tag(self, tag: str) -> None:
        """ Add a tag to this nonterminal """
        if self._tags is None:
            self._tags = {tag}
        else:
            self._tags.add(tag)
        # Optimization for the no_reduce tag
        if tag == "no_reduce":
            self.no_reduce = True

    @property
    def is_noun_phrase(self) -> bool:
        """ This is overwritten in BIN_Nonterminal in binparser.py """
        return False

    def __repr__(self) -> str:
        return self._name

    def __str__(self) -> str:
        return self._name


class Terminal(GrammarItem):

    """ A terminal within a right-hand-side production """

    _index = 1  # Running sequence number (positive) of all terminals

    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name
        self._index = Terminal._index
        Terminal._index += 1
        # The hash is used quite often so it is worth caching
        self._hash = id(self).__hash__()

    def __hash__(self) -> int:
        return self._hash

    def __repr__(self) -> str:
        return self._name

    def __str__(self) -> str:
        return self._name

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_literal(self) -> bool:
        """ Return True if this is a literal terminal, i.e. one enclosed
            within single or double quotes """
        return False

    def set_index(self, ix: int) -> None:
        """ Set a new sequence number for this nonterminal """
        # Terminals have indices larger than zero
        assert ix > 0
        super().set_index(ix)

    def matches(self, t_kind: str, t_val: str, t_lit: str) -> bool:
        """ Does this terminal match the given token? """
        return self._name == t_kind


class LiteralTerminal(Terminal):

    """ A literal (constant string) terminal within a right-hand-side production.
        A literal within single quotes 'x' is matched canonically, i.e. with
        the corresponding word stem, if available.
        A literal within double quotes "x" is matched absolutely, i.e. with
        the exact source text (except for a conversion to lowercase). """

    def __init__(self, lit: str) -> None:
        q = lit[0]
        assert q in "'\""
        super().__init__(lit)
        # If a double quote was used, this is a 'strong' literal
        # that matches an exact terminal string as it appeared in the source
        # - no stemming or other canonization should be applied,
        # although the string will be converted to lowercase
        self._strong = q == '"'

    @property
    def literal_text(self) -> str:
        """ If this is a strong literal, return that literal """
        if self._strong:
            return self._name[1:-1]
        return ""

    @property
    def is_literal(self) -> bool:
        """ Return True if this is a literal terminal, i.e. one enclosed
            within single or double quotes """
        return True

    def matches(self, t_kind: str, t_val: str, t_lit: str) -> bool:
        """ A literal terminal matches a token if the token text is
            canonically or absolutely identical to the literal """
        if self._strong:
            # Absolute literal match
            return self._name == t_lit
        # Canonical match of stems or prototypes
        return self._name == t_val


class Token:

    """ A single input token as seen by the parser """

    def __init__(self, kind: str, val: str, lit: Optional[str] = None) -> None:
        """ A basic token has a kind, a canonical value
            and an optional literal value, all strings """
        self._kind = kind
        self._val = val
        self._lit = lit or val

    def __repr__(self) -> str:
        """ Return a simple string representation of this token """
        if self._kind == self._val:
            return self._kind
        return "{0}:{1}".format(self._kind, self._val)

    @property
    def kind(self) -> str:
        """ Return the token kind """
        return self._kind

    @property
    def text(self) -> str:
        """ Return the 'canonical' token text, which may be a stem or
            prototype of the literal, original token text as it appeared
            in the source """
        return self._val

    @property
    def literal(self) -> str:
        """ Return the literal, original token text as it appeared in the source """
        return self._lit

    def matches(self, terminal: Terminal) -> bool:
        """ Does this token match the given terminal? """
        # By default, ask the terminal
        return terminal.matches(self._kind, self._val, self._lit)


class Production:

    """ A right-hand side of a grammar rule """

    _index = 0  # Running sequence number of all productions

    def __init__(
        self,
        fname: Optional[str] = None,
        line: int = 0,
        rhs: Optional[List[GrammarItem]] = None,
        priority: int = 0,
    ) -> None:

        """ Initialize a production from a list of
            right-hand-side nonterminals and terminals """

        self._rhs: List[GrammarItem] = [] if rhs is None else rhs
        # If parsing a grammar file, note the position of the production
        # in the file
        self._fname = fname
        self._line = line
        self._priority = priority
        # Cache the length of the production as it is used A LOT
        self._len = len(self._rhs)
        # Give all productions a unique sequence number for hashing purposes
        self._index = Production._index
        Production._index += 1
        # Cached tuple representation of this production
        self._tuple: Optional[Tuple[int, ...]] = None

    @classmethod
    def reset(cls) -> None:
        """ Reset the production index sequence to zero """
        cls._index = 0

    def append(self, t: GrammarItem) -> None:
        """ Append a terminal or nonterminal to this production """
        self._rhs.append(t)
        self._len += 1
        # Destroy the cached tuple, if any
        self._tuple = None

    @property
    def index(self) -> int:
        return self._index

    @property
    def length(self) -> int:
        """ Return the length of this production """
        return self._len

    @property
    def is_empty(self) -> bool:
        """ Return True if this is an empty (epsilon) production """
        return self._len == 0

    @property
    def fname(self) -> Optional[str]:
        return self._fname

    @property
    def line(self) -> int:
        return self._line

    @property
    def priority(self) -> int:
        return self._priority

    @property
    def prod(self) -> Tuple[int, ...]:
        """ Return this production in tuple form """
        if self._tuple is None:
            # Nonterminals have negative indices and terminals have positive ones
            self._tuple = tuple(t.index for t in self._rhs) if self._rhs else tuple()
        return self._tuple

    def nonterminal_at(self, dot: int) -> bool:
        """ Return True if prod[dot] is a nonterminal or completed """
        return dot >= self._len or isinstance(self._rhs[dot], Nonterminal)

    def __getitem__(self, index: int) -> GrammarItem:
        """ Return the Terminal or Nonterminal at the given index position """
        return self._rhs[index]

    def __setitem__(self, index: int, item: GrammarItem):
        """ Set the Terminal or Nonterminal at the given index position """
        self._rhs[index] = item

    def __len__(self) -> int:
        """ Return the length of this production """
        return self._len

    def __iter__(self) -> Iterator[GrammarItem]:
        return iter(self._rhs)

    def __repr__(self) -> str:
        """ Return a representation of this production """
        return "<P: " + repr(self._rhs) + ">"

    def __str__(self) -> str:
        """ Return a representation of this production """
        return " ".join([str(t) for t in self._rhs]) if self._rhs else "0"


_PRAGMA_SCORE = "$score("
_PRAGMA_ROOT = "$root("
_PRAGMA_TAG = "$tag("
_PRAGMA_IF = "$if("
_PRAGMA_ENDIF = "$endif("


class Grammar:

    """
        A grammar maps nonterminals to a list of right hand sides.
        Each right hand side is a list of terminals and nonterminals.

        The text representation of a grammar is as follows:

        A -> A B terminal C
            | A '/' D
            | 0
        B -> terminal "+" C

        Nonterminals start with uppercase letters.

        Terminals start with lowercase letters or are enclosed
        in single or double quotes.

        0 means an empty (epsilon) production.

        As an alternative to the '|' symbol, productions may also
        be separated by '>'. This indicates a priority ordering
        between the alternatives, so that the result tree is
        automatically reduced to the highest-priority alternative
        in case of ambiguity.

    """

    def __init__(self) -> None:

        self._nonterminals: Dict[str, Nonterminal] = OrderedDict()
        self._terminals: Dict[str, Terminal] = OrderedDict()

        # Dictionary of nonterminals indexed by integers < 0
        self._nonterminals_by_ix: Dict[int, Nonterminal] = {}
        # Dictionary of terminals indexed by integers > 0
        self._terminals_by_ix: Dict[int, Terminal] = {}
        # Dictionary of productions indexed by integers >= 0
        self._productions_by_ix: Dict[int, Production] = {}

        # Mapping of nonterminals to a list of their productions
        self._nt_dict: Dict[Nonterminal, List[ProductionTuple]] = OrderedDict()
        # Nonterminal score adjustment as specified by $score(n)
        self._nt_scores: Dict[Nonterminal, int] = {}

        self._root: Optional[Nonterminal] = None
        # Additional, secondary roots, if any
        self._secondary_roots: List[Nonterminal] = []

        # Information about the grammar file
        self._file_name: Optional[str] = None
        self._file_time: Optional[datetime] = None

        # Grammar parsing conditions, checkable with $if()...$endif()
        # This should be a set of strings
        self._conditions: Set[str] = set()

    @property
    def nt_dict(self) -> Dict[Nonterminal, List[ProductionTuple]]:
        """ Return the raw grammar dictionary, Nonterminal -> [ Productions ] """
        return self._nt_dict

    def nt_score(self, nt: Nonterminal) -> int:
        """ Return the score adjustment for the given nonterminal """
        return self._nt_scores.get(nt, 0)

    @property
    def root(self) -> Optional[Nonterminal]:
        """ Return the root nonterminal for this grammar """
        return self._root

    @property
    def terminals(self) -> Dict[str, Terminal]:
        """ Return a dictionary of terminals in the grammar """
        return self._terminals

    @property
    def nonterminals(self) -> Dict[str, Nonterminal]:
        """ Return a dictionary of nonterminals in the grammar """
        return self._nonterminals

    @property
    def nonterminals_by_ix(self) -> Dict[int, Nonterminal]:
        """ Return a dictionary of nonterminals in the grammar,
            indexed by integer < 0 """
        return self._nonterminals_by_ix

    @property
    def terminals_by_ix(self) -> Dict[int, Terminal]:
        """ Return a dictionary of terminals in the grammar,
            indexed by integer > 0 """
        return self._terminals_by_ix

    @property
    def productions_by_ix(self) -> Dict[int, Production]:
        """ Return a dictionary of productions in the grammar,
            indexed by integer >= 0 """
        return self._productions_by_ix

    def lookup(self, index: int) -> Optional[GrammarItem]:
        """ Look up a nonterminal or terminal by integer index """
        if index < 0:
            return self._nonterminals_by_ix.get(index)
        if index > 0:
            return self._terminals_by_ix.get(index)
        # index == 0
        return None

    def lookup_nonterminal(self, index: int) -> Optional[Nonterminal]:
        """ Look up a nonterminal by (negative) integer index """
        return None if index >= 0 else self._nonterminals_by_ix.get(index)

    def lookup_terminal(self, index: int) -> Optional[Terminal]:
        """ Look up a terminal by (positive) integer index """
        return None if index <= 0 else self._terminals_by_ix.get(index)

    @property
    def num_nonterminals(self) -> int:
        """ Return the number of nonterminals in the grammar """
        return len(self._nonterminals)

    @property
    def num_terminals(self) -> int:
        """ Return the number of terminals in the grammar """
        return len(self._terminals)

    @property
    def num_productions(self) -> int:
        """ Return the total number of productions in the grammar,
            were each right hand side option is counted as one """
        return sum(len(pp) for pp in self._nt_dict.values())

    @property
    def file_name(self) -> Optional[str]:
        """ Return the name of the grammar file, or None """
        return self._file_name

    @property
    def file_time(self) -> Optional[datetime]:
        """ Return the timestamp of the grammar file, or None """
        return self._file_time

    def set_conditions(self, cond_set: Set[str]) -> None:
        """ Set the parsing conditions for this grammar,
            checkable with $if()...$endif() """
        self._conditions = cond_set

    def __getitem__(self, nt: Nonterminal) -> List[ProductionTuple]:
        """ Look up a nonterminal, yielding a list of (priority, production) tuples """
        return self._nt_dict[nt]

    def __str__(self) -> str:
        """ Return a string representation of this grammar """

        def to_str(plist: Iterable[ProductionTuple]) -> str:
            return " | ".join([str(p) for _, p in plist])

        return "\n".join(
            [str(nt) + " → " + to_str(ptl) + "\n" for nt, ptl in self._nt_dict.items()]
        )

    @staticmethod
    def _make_terminal(name: str) -> Terminal:
        """ Create a new Terminal instance within the grammar """
        # Override this to create custom terminals or add optimizations
        return Terminal(name)

    @staticmethod
    def _make_literal_terminal(name: str) -> LiteralTerminal:
        """ Create a new LiteralTerminal instance within the grammar """
        # Override this to create custom terminals or add optimizations
        return LiteralTerminal(name)

    @staticmethod
    def _make_nonterminal(name: str, fname: str, line: int) -> Nonterminal:
        """ Create a new Nonterminal instance within the grammar """
        # Override this to create custom nonterminals or add optimizations
        return Nonterminal(name, fname, line)

    def _write_binary(self, fname: str) -> None:
        """ Write grammar to binary file. Called after reading a grammar text file
            that is newer than the corresponding binary file. """
        with open(fname, "wb") as f:
            if Settings.DEBUG:
                print("Writing binary grammar file {0}".format(fname))
            # Version header
            f.write("Greynir00.00.01\n".encode("ascii"))  # 16 bytes total
            num_nt = self.num_nonterminals
            # Number of terminals and nonterminals in grammar
            f.write(struct.pack("<II", self.num_terminals, num_nt))
            # Root nonterminal
            assert self.root is not None
            if Settings.DEBUG:
                print("Root index is {0}".format(self.root.index))
            f.write(struct.pack("<i", self.root.index))
            # Write nonterminals in numeric order, -1 first downto -N
            for ix in range(num_nt):
                nt = self.lookup_nonterminal(-1 - ix)
                plist = self[nt] if nt else []
                f.write(struct.pack("<I", len(plist)))
                # Write productions along with their indices and priorities
                for prio, p in plist:
                    lenp = len(p)
                    f.write(struct.pack("<III", p.index, prio, lenp))
                    if lenp:
                        f.write(struct.pack("<" + str(lenp) + "i", *p.prod))
        if Settings.DEBUG:
            print("Writing of binary grammar file completed")
            print(
                "num_terminals was {0}, num_nonterminals {1}".format(
                    self.num_terminals, num_nt
                )
            )

    def read(
        self, fname: str, verbose: bool = False, binary_fname: Optional[str] = None
    ) -> None:
        """ Read grammar from a text file. Set verbose=True to get diagnostic messages
            about unused nonterminals and nonterminals that are unreachable
            from the root.
            Pass a file name in binary_fname to write a fresh binary file if the
            grammar text file is newer than the existing binary file. """
        try:
            with open(fname, "r", encoding="utf-8") as inp:
                # Read grammar file line-by-line
                self.read_from_generator(fname, inp, verbose, binary_fname)
        except (IOError, OSError):
            raise GrammarError("Unable to open or read grammar file", fname, 0)

    def read_from_generator(
        self,
        fname: str,
        line_generator: Iterator[str],
        verbose: bool = False,
        binary_fname: Optional[str] = None,
        force_new_binary: bool = False,
    ) -> None:
        """ Read grammar from a generator of lines. Set verbose=True to get
            diagnostic messages about unused nonterminals and nonterminals that are
            unreachable from the root. Pass a file name in binary_fname to write
            a fresh binary file if the grammar text file is newer than the
            existing binary file. Set force_new_binary=True to write a fresh
            binary file regardless of timestamps. """

        # Clear previous file info, if any
        self._file_time = self._file_name = None
        # Shortcuts
        terminals = self._terminals
        nonterminals = self._nonterminals
        grammar: Dict[Nonterminal, List[ProductionTuple]] = self._nt_dict
        # Reset the sequence of production indices
        Production.reset()
        # Dictionary of variants, keyed by variant name
        # where the values are lists of variant options (strings)
        variants: Dict[str, List[str]] = OrderedDict()

        def parse_line(s: str) -> None:

            s = s.strip()
            if not s:
                # Blank line: ignore
                return

            def highest_priority(nt_id: str) -> int:
                """ Return the highest priority index for a production
                    of the given nonterminal """
                try:
                    nt = nonterminals[nt_id]
                    return max(prio for prio, _ in grammar[nt])
                except (KeyError, ValueError):
                    return 0

            def parse_rhs(nt_id: str, vts: List[str], s: str, priority: int):
                """ Parse a right-hand side sequence, eventually with
                    relative priority within the nonterminal """

                def add_rhs(
                    nt_id: str, rhs: Optional[Production], priority: int = 0
                ) -> None:
                    """ Add a fully expanded right-hand-side production
                        to a nonterminal rule """
                    nt = nonterminals[nt_id]
                    if nt not in grammar:
                        # First production of this nonterminal
                        grammar[nt] = [] if rhs is None else [(priority, rhs)]
                        return
                    if rhs is None:
                        return
                    if rhs.is_empty:
                        # Adding epsilon production: avoid multiple ones
                        if any(p.is_empty for _, p in grammar[nt]):
                            return
                    # Append to the list of productions of this nonterminal
                    grammar[nt].append((priority, rhs))

                s = s.strip()
                if not s:
                    raise GrammarError("Invalid syntax for production", fname, line)

                tokens = s.split()

                # rhs is a list of tuples, one for each token, as follows:
                # (id, repeat, variants)
                rhs: List[Tuple[Optional[str], Optional[str], Optional[List[str]]]] = []

                # vfree is a set of 'free variants', i.e. variants that
                # occur in the right hand side of the production but not in
                # the nonterminal (those are in vts)
                vfree: Set[str] = set()

                for r in tokens:

                    if r in {"0", "ø", "∅"}:
                        # Empty (epsilon) production
                        if len(tokens) != 1:
                            raise GrammarError(
                                "Empty (epsilon) rule must be of the form NT -> 0",
                                fname,
                                line,
                            )
                        rhs.append((None, None, None))
                        break

                    # Check for repeat/conditionality
                    repeat = None
                    if r[-1] in "*+?":
                        # Optional repeat/conditionality specifier
                        # Asterisk: Can be repeated 0 or more times
                        # Plus: Can be repeated 1 or more times
                        # Question mark: optionally present once
                        repeat = r[-1]
                        r = r[0:-1]

                    # Check for variant specs
                    if r == "\"/\"" or r == "'/'":
                        # Hack: Allow specifying the forward slash as "/" or '/'
                        rsplit = [r]
                    else:
                        rsplit = r.split("/")
                        r = rsplit[0]
                    v: Optional[List[str]] = rsplit[1:]
                    if not v:
                        v = None
                    else:
                        for vspec in v:
                            if vspec not in variants:
                                raise GrammarError(
                                    "Unknown variant '{0}'".format(vspec), fname, line
                                )
                            if vspec not in vts:
                                # Free variant: add to set
                                vfree.add(vspec)

                    if r[0] in "\"'":
                        # Literal terminal symbol
                        if r == '""':
                            # Empty literal: matches nothing
                            pass
                        elif len(r) < 3 or r[0] not in r[2:]:
                            raise GrammarError(
                                "Invalid literal terminal {0}".format(r), fname, line
                            )
                    else:
                        # Identifier of nonterminal or terminal
                        if not r.isidentifier():
                            raise GrammarError(
                                "Invalid identifier '{0}'".format(r), fname, line
                            )
                    rhs.append((r, repeat, v))

                assert len(rhs) == len(tokens)

                # Generate productions for all variants

                def variant_values(vlist: List[str]) -> Iterator[List[str]]:
                    """ Returns a list of names with all applicable
                        variant options appended """
                    if not vlist:
                        yield [""]
                        return
                    if len(vlist) == 1:
                        for vopt in variants[vlist[0]]:
                            yield [vopt]
                        return
                    for v in variant_values(vlist[1:]):
                        for vopt in variants[vlist[0]]:
                            yield [vopt] + v

                # Make a list of all variants that occur in the
                # nonterminal or on the right hand side
                vall = vts + list(vfree)

                for vval in variant_values(vall):
                    # Generate a production for every variant combination

                    # Calculate the nonterminal suffix for this variant
                    # combination
                    nt_suffix = (
                        "_".join(vval[vall.index(vx)] for vx in vts) if vts else ""
                    )
                    if nt_suffix:
                        nt_suffix = "_" + nt_suffix

                    result = Production(fname, line, priority=priority)
                    for rname, repeat, v in rhs:
                        # Calculate the token suffix, if any
                        # This may be different from the nonterminal suffix as
                        # the token may have fewer variants than the nonterminal,
                        # and/or free ones that don't appear in the nonterminal.
                        n: Union[None, Nonterminal, Terminal]
                        if rname is None:
                            # Epsilon
                            n = None
                            sym = ""
                        else:
                            suffix = (
                                "_".join(vval[vall.index(vx)] for vx in v) if v else ""
                            )
                            if suffix:
                                suffix = "_" + suffix
                            sym = rname + suffix
                            if rname[0] in "'\"":
                                # Literal terminal
                                if sym not in terminals:
                                    terminals[sym] = self._make_literal_terminal(sym)
                                n = terminals[sym]
                            elif rname[0].isupper():
                                # Identifier of nonterminal
                                if sym not in nonterminals:
                                    nonterminals[sym] = self._make_nonterminal(
                                        sym, fname, line
                                    )
                                n = nonterminals[sym]
                                # Note that the nonterminal has been referenced
                                n.add_ref()
                            else:
                                # Identifier of terminal
                                if sym not in terminals:
                                    terminals[sym] = self._make_terminal(sym)
                                n = terminals[sym]

                        # Convert Enhanced Backus-Naur Format (EBNF)
                        # to plain old BNF

                        # If the production item can be repeated or is optional,
                        # create a new production and substitute.

                        # A -> B C* D becomes:
                        # A -> B C_star D
                        # C_star -> C_star C | 0

                        # A -> B C+ D becomes:
                        # A -> B C_plus D
                        # C_plus -> C_plus C | C

                        # A -> B C? D becomes:
                        # A -> B C_opt D
                        # C_opt -> C | 0

                        # (The new productions are actually called
                        # C*, C+ and C? respectively, not C_star, C_plus, C_opt)

                        if repeat is not None:
                            if n is None:
                                raise GrammarError(
                                    "Epsilon (0) cannot be suffixed with ?, * or +",
                                    fname,
                                    line,
                                )
                            # Create C*, C+ or C?
                            new_nt_id = sym + repeat
                            # Make the new nonterminal and production,
                            # if not already there
                            if new_nt_id not in nonterminals:
                                new_nt = nonterminals[
                                    new_nt_id
                                ] = self._make_nonterminal(new_nt_id, fname, line)
                                new_nt.add_ref()
                                # Note that the Earley algorithm is more efficient
                                # on left recursion than middle or right recursion.
                                # Therefore it is better to generate
                                # Cx -> Cx C than Cx -> C Cx.
                                # First production: Cx C
                                new_p = Production(fname, line)
                                if repeat != "?":
                                    new_p.append(new_nt)  # C* / C+
                                new_p.append(n)  # C
                                add_rhs(new_nt_id, new_p)  # Default priority 0
                                # Second production: epsilon(*, ?) or C(+)
                                new_p = Production(fname, line)
                                if repeat == "+":
                                    new_p.append(n)
                                add_rhs(new_nt_id, new_p)  # Default priority 0
                            # Substitute the Cx in the original production
                            n = nonterminals[new_nt_id]

                        if n is not None:
                            result.append(n)

                    assert len(result) == len(rhs) or (
                        len(rhs) == 1 and rhs[0] == (None, None, None)
                    )

                    nt_id_full = nt_id + nt_suffix

                    if len(result) == 1 and result[0] == nonterminals[nt_id_full]:
                        # Nonterminal derives itself
                        raise GrammarError(
                            "Nonterminal {0} deriving itself".format(nt_id_full),
                            fname,
                            line,
                        )
                    add_rhs(nt_id_full, result, priority)

            def variant_names(nt: str, vts: List[str]) -> List[str]:
                """ Returns a list of names with all applicable
                    variant options appended """
                result = [nt]
                for v in vts:
                    newresult: List[str] = []
                    for vopt in variants[v]:
                        for r in result:
                            newresult.append(r + "_" + vopt)
                    result = newresult
                return result

            def apply_to_nonterminals(s: str, func: Callable[[Nonterminal, str], None]) -> None:
                """ Parse a nonterminal/var list from string s,
                    then apply func(nt, p) to all nonterminals,
                    where p is the parameter of the pragma """
                ix = s.rfind(")")
                if ix < 0:
                    raise GrammarError(
                        "Expected right parenthesis in pragma", fname, line
                    )
                param = s[0:ix].strip()
                s = s[ix + 1 :]
                nts = s.split()
                cnt = 0
                for nt_name in nts:
                    ntv = nt_name.split("/")
                    for vname in ntv[1:]:
                        if vname not in variants:
                            raise GrammarError(
                                "Unknown variant '{0}' for nonterminal '{1}'".format(
                                    vname, ntv[0]
                                ),
                                fname,
                                line,
                            )
                    for vname in variant_names(ntv[0], ntv[1:]):
                        if vname not in nonterminals:
                            raise GrammarError(
                                "Unknown nonterminal '{0}'".format(vname), fname, line
                            )
                        try:
                            # Found a nonterminal to which this pragma applies
                            func(nonterminals[vname], param)
                            cnt += 1
                        except:
                            raise GrammarError(
                                "Invalid pragma argument '{0}'".format(param),
                                fname,
                                line,
                            )
                if cnt == 0:
                    # This pragma has no effect
                    raise GrammarError(
                        "Pragma does not affect any nonterminal", fname, line
                    )

            if s.startswith("/"):

                # Definition of variant
                # A variant is specified as /varname = opt1 opt2 opt3...
                v = s.split("=", maxsplit=1)
                if len(v) != 2:
                    raise GrammarError("Invalid variant syntax", fname, line)
                vname = v[0].strip()[1:]
                if "_" in vname or not vname.isidentifier():
                    # Variant names must be valid identifiers without underscores
                    raise GrammarError(
                        "Invalid variant name '{0}'".format(vname), fname, line
                    )
                v = v[1].split()
                for vopt in v:
                    if "_" in vopt or not vopt.isidentifier():
                        # Variant options must be valid identifiers without underscores
                        raise GrammarError(
                            "Invalid option '{0}' in variant '{1}'".format(vopt, vname),
                            fname,
                            line,
                        )
                variants[vname] = v

            elif s.startswith("$"):

                # Pragma
                s = s.strip()

                if s.startswith(_PRAGMA_SCORE):

                    # Pragma $score(int) Nonterminal/var1/var2 ...
                    s = s[len(_PRAGMA_SCORE) :]

                    def set_score(nt: Nonterminal, score: str) -> None:
                        self._nt_scores[nt] = int(score)

                    apply_to_nonterminals(s, set_score)

                elif s.startswith(_PRAGMA_TAG):

                    # Pragma $tag(tagstring) Nonterminal/var1/var2 ...
                    s = s[len(_PRAGMA_TAG) :]
                    apply_to_nonterminals(s, lambda nt, tag: nt.add_tag(tag))

                elif s.startswith(_PRAGMA_ROOT):

                    # Pragma $root(Nonterminal)
                    # Identify a nonterminal as a secondary parse root
                    if s[-1] != ")":
                        raise GrammarError(
                            "Expected right parenthesis in $root() pragma", fname, line
                        )
                    root_nt = s[len(_PRAGMA_ROOT) : -1].strip()
                    if not root_nt.isidentifier():
                        raise GrammarError(
                            "Invalid nonterminal name '{0}'".format(root_nt),
                            fname,
                            line,
                        )
                    if root_nt not in nonterminals:
                        raise GrammarError("Unknown nonterminal '{0}'".format(root_nt))
                    # Add an implicit reference to the root
                    nonterminals[root_nt].add_ref()
                    self._secondary_roots.append(nonterminals[root_nt])

                else:
                    raise GrammarError("Unknown pragma '{0}'".format(s), fname, line)

            else:

                # New nonterminal
                if "→" in s:
                    # Fancy schmancy arrow sign: use it
                    rule = s.split("→", maxsplit=1)
                else:
                    rule = s.split("->", maxsplit=1)
                if len(rule) != 2:
                    raise GrammarError("Invalid syntax", fname, line)

                # Split nonterminal spec into name and variant(s),
                # i.e. NtName/var1/var2...
                ntv = rule[0].strip().split("/")
                current_NT = nt = ntv[0]
                current_variants = ntv[1:]
                if not nt.isidentifier():
                    raise GrammarError(
                        "Invalid nonterminal name '{0}'".format(nt), fname, line
                    )
                for vname in current_variants:
                    if vname not in variants:
                        raise GrammarError(
                            "Unknown variant '{0}' for nonterminal '{1}'".format(
                                vname, nt
                            ),
                            fname,
                            line,
                        )

                # Add all previously unknown nonterminal variants
                for nt_var in variant_names(nt, current_variants):
                    cnt: Nonterminal
                    if nt_var in nonterminals:
                        cnt = nonterminals[nt_var]
                    else:
                        cnt = self._make_nonterminal(nt_var, fname, line)
                        nonterminals[nt_var] = cnt
                        if self._root is None:
                            # Remember first nonterminal as the root
                            self._root = cnt
                            self._root.add_ref()  # Implicitly referenced
                    if cnt not in grammar:
                        grammar[cnt] = []

                sep = "|"  # Default production separator
                priority = 0  # Default production priority
                if ">" in rule[1]:
                    # Looks like a priority specification between productions
                    if "|" in rule[1]:
                        raise GrammarError(
                            "Cannot mix '|' and '>' between productions", fname, line
                        )
                    sep = ">"
                    # In the case of a prioritized list of productions,
                    # start with a priority index equal to the highest priority
                    # of the existing productions plus one, ensuring that
                    # additional productions always have less priority
                    # (higher index) than the pre-existing ones
                    priority = highest_priority(current_NT) + 1

                for prod in rule[1].split(sep):
                    # Add the productions on the right hand side,
                    # delimited by '|' or '>'
                    # Note: we explicitly tolerate extra separators without error;
                    # this can be used to attach lower-priority productions
                    # to nonterminals, for instance fallback rules to be used
                    # for grammar error checking.
                    prod = prod.strip()
                    if prod:
                        parse_rhs(current_NT, current_variants, prod, priority)
                        # If productions are separated by the > sign, we increase
                        # the priority number (which lowers the priority) by one
                        # for each production
                        if sep == ">":
                            priority += 1

        # Main parse loop

        # The number of the current line in the grammar file
        line = 0
        # The current logical line, eventually amalgamated from
        # multiple physical lines via continuation
        current_line = ""
        # Stack of conditional sections
        cond_stack: List[Tuple[str, bool]] = [("", True)]

        try:
            # Read grammar file line-by-line

            for s in line_generator:

                line += 1
                # Ignore comments
                ix = s.find("#")
                if ix >= 0:
                    s = s[0:ix]

                s = s.rstrip()
                if not s:
                    continue

                # If line starts with a blank, assume it's a continuation
                if s[0].isspace():
                    current_line += s
                    continue

                if cond_stack[-1][1]:
                    # New item starting: parse the previous one and start a new
                    parse_line(current_line)

                # Check conditional section
                if s.startswith(_PRAGMA_IF) and s.endswith(")"):
                    cond = s[4:-1].strip()
                    if not cond.isidentifier():
                        raise GrammarError("$if() condition must be a valid identifier")
                    # Establish whether we want to parse the conditional
                    # section or not
                    new_cond = cond_stack[-1][1] and cond in self._conditions
                    cond_stack.append((cond, new_cond))
                    s = ""
                elif s.startswith(_PRAGMA_ENDIF) and s.endswith(")"):
                    cond = s[7:-1].strip()
                    if not cond.isidentifier():
                        raise GrammarError(
                            "$endif() condition must be a valid identifier"
                        )
                    if len(cond_stack) < 2:
                        raise GrammarError("$endif() with no matching $if()")
                    if cond != cond_stack[-1][0]:
                        raise GrammarError(
                            "$endif({0}) does not match $if({1})".format(
                                cond, cond_stack[-1][0]
                            )
                        )
                    del cond_stack[-1]
                    s = ""

                current_line = s

            # Parse the final chunk
            if current_line:
                parse_line(current_line)

        except GrammarError as e:
            e.augment(fname, line)
            raise e

        # Check all nonterminals to verify that they have productions and are referenced
        for nt in nonterminals.values():
            if verbose and not nt.has_ref:
                # Emit a warning message if verbose=True
                print("Nonterminal {0} is never referenced in a production".format(nt))
            if nt not in grammar:
                raise GrammarError(
                    "Nonterminal {0} is referenced but not defined".format(nt),
                    nt.fname,
                    nt.line,
                )
        for nt, plist in grammar.items():
            if len(plist) == 0:
                raise GrammarError(
                    "Nonterminal {0} has no productions".format(nt), nt.fname, nt.line
                )
            else:
                for _, p in plist:
                    if len(p) == 1 and p[0] == nt:
                        raise GrammarError(
                            "Nonterminal {0} produces itself".format(nt),
                            p.fname,
                            p.line,
                        )

        # Check that productions do not contain consecutive literals
        # that are joined to become a single multi-word phrase token
        # Loop through nonterminals of the grammar
        for _, plist in grammar.items():
            # Loop through the production list of a nonterminal
            for _, p in plist:
                phrase: List[str] = []
                # Loop through items in a production
                for item in p:
                    item_text = item.literal_text
                    if item_text:
                        # This is a literal terminal: append to phrase
                        phrase.append(item_text)
                        if len(phrase) >= 2:
                            # We have a phrase: check it
                            ptxt = " ".join(phrase)
                            if ptxt in StaticPhrases.MAP:
                                # Oops: defined as a multi-word token in
                                # the [static_phrases] section of Phrases.conf
                                raise GrammarError(
                                    "Consecutive literal terminals match "
                                    "static phrase; use {0} instead".format(
                                        '"' + "_".join(phrase) + '"'
                                    )
                                )
                    else:
                        # Restart phrase
                        phrase = []

        # Check that all nonterminals derive terminal strings
        agenda = [nt for nt in nonterminals.values()]
        der_t: Set[Nonterminal] = set()
        while agenda:
            reduced = False
            for nt in agenda:
                for _, p in grammar[nt]:
                    if all(isinstance(s, Terminal) or s in der_t for s in p):
                        der_t.add(nt)
                        break
                if nt in der_t:
                    reduced = True
            if not reduced:
                break
            agenda = [nt for nt in nonterminals.values() if nt not in der_t]
        if agenda:
            raise GrammarError(
                "Nonterminals {0} do not derive terminal strings".format(
                    ", ".join([str(nt) for nt in agenda])
                ),
                fname,
                0,
            )

        # Short-circuit nonterminals that point directly and uniquely
        # to other nonterminals.
        # Becausee this creates a gap between the original grammar
        # and the resulting trees, we only do this for nonterminals with variants
        # that do not have a $score pragma

        # Dictionary of shortcuts
        shortcuts: Dict[Nonterminal, Nonterminal] = {}

        for nt, plist in grammar.items():
            if not "_" in nt.name:
                # 'Pure' nonterminal with no variants: don't shortcut
                continue
            if self.nt_score(nt) != 0 or nt.has_tags:
                # Nonterminal has a score adjustment or a tag: don't shortcut
                continue
            if (
                len(plist) == 1
                and len(plist[0][1]) == 1
                and isinstance(plist[0][1][0], Nonterminal)
            ):
                # This nonterminal has only one production,
                # with only one nonterminal item
                target_nt: Nonterminal = cast(Nonterminal, plist[0][1][0])
                assert target_nt != nt
                while target_nt in shortcuts:
                    # Find ultimate destination of shortcut
                    assert target_nt != shortcuts[target_nt]
                    target = shortcuts[target_nt]
                shortcuts[nt] = target_nt

        # Go through all productions and replace the shortcuts with their targets
        for nt, plist in grammar.items():
            for _, p in plist:
                for ix, item in enumerate(p):
                    if isinstance(item, Nonterminal) and item in shortcuts:
                        # Replace the nonterminal in the production
                        target = shortcuts[item]
                        # if verbose:
                        #    # Print informational message in verbose mode
                        #    print("Production of {2}: Replaced {0} with {1}"
                        #        .format(s, target, nt))
                        p[ix] = target

        # Now, after applying shortcuts, check that all nonterminals
        # are reachable from the root
        unreachable = {nt for nt in nonterminals.values()}

        def _remove(nt: Nonterminal) -> None:
            """ Recursively remove all nonterminals that are reachable from nt """
            unreachable.remove(nt)
            for _, p in grammar[nt]:
                for s in p:
                    if isinstance(s, Nonterminal) and s in unreachable:
                        _remove(s)

        # Remove the main root and any secondary roots
        if self._root is not None:
            _remove(self._root)
            for r in self._secondary_roots:
                _remove(r)

        if unreachable:
            if verbose:
                # Emit a warning message if verbose=True
                print(
                    "The following nonterminals are unreachable from the root\n"
                    "and will be removed from the grammar:"
                )
                with changedlocale() as strxfrm:
                    for nt_name in sorted([str(nt) for nt in unreachable], key=strxfrm):
                        print("* {0}".format(nt_name))
            # Simplify the grammar dictionary by removing unreachable nonterminals
            for nt in unreachable:
                del grammar[nt]
                del nonterminals[nt.name]

        # Reassign indices for nonterminals to avoid gaps in the number sequence
        # Nonterminals are indexed downwards from -1
        # We must take care to sort the dictionary before enumerating it,
        # as the ordering will otherwise be non-deterministic and non-repeatable.
        # Note also that the sorting is by design not done on the Icelandic locale,
        # but rather on the default Python string sorting order for maximum
        # repeatability and to avoid requiring the is-IS locale to be installed.
        nt_sorted = sorted(nonterminals.keys())
        self._nonterminals_by_ix = {
            -1 - ix: nonterminals[key] for ix, key in enumerate(nt_sorted)
        }
        for key, nt in self._nonterminals_by_ix.items():
            nt.set_index(key)

        # Reassign indices for terminals
        # Terminals are indexed upwards from 1
        t_sorted = sorted(terminals.keys())
        self._terminals_by_ix = {
            ix + 1: terminals[key] for ix, key in enumerate(t_sorted)
        }
        for key, t in self._terminals_by_ix.items():
            t.set_index(key)

        # Make a dictionary of productions by integer index >= 0.
        # Here, no sorting is required for repeatability, as global production indices
        # are assigned by order of appearance in the grammar file, and the same
        # applies for the local order of productions within a nonterminal.
        for plist in grammar.values():
            for _, p in plist:
                self._productions_by_ix[p.index] = p

        # Grammar successfully read: note the file name and timestamp
        self._file_name = fname
        self._file_time = datetime.fromtimestamp(os.path.getmtime(fname))

        if binary_fname is not None:
            # Check whether to write a fresh binary file
            if force_new_binary:
                # Always write a fresh binary file if this flag is set
                binary_file_time = None
            else:
                try:
                    binary_file_time = datetime.fromtimestamp(
                        os.path.getmtime(binary_fname)
                    )
                except os.error:
                    binary_file_time = None
            if binary_file_time is None or binary_file_time < self._file_time:
                # No binary file or older than text file: write a fresh one
                self._write_binary(binary_fname)

    def follow_set(self, nonterminal: Nonterminal) -> Dict[Terminal, List[List[Production]]]:
        """ Return the set of terminals that can follow
            the given nonterminal, as a dictionary keyed
            by terminal, containing a list of productions
            by which the terminal follows the nonterminal """

        nullable: Set[Nonterminal] = set()

        def is_nullable(nt: Nonterminal) -> bool:
            """ Returns True if the nonterminal is nullable """

            def is_nullable_prod(p: Production) -> bool:
                """ Returns True if the production is nullable """
                return p.is_empty or all(s in nullable for s in p)

            plist = self._nt_dict[nt]
            return any(is_nullable_prod(p) for _, p in plist)

        changed = True
        while changed:
            changed = False
            for nt in self._nonterminals.values():
                if nt not in nullable and is_nullable(nt):
                    nullable.add(nt)
                    changed = True

        follow: DefaultDict[Terminal, List[List[Production]]] = defaultdict(list)
        seen: Set[GrammarItem] = set()

        def add_follow(nt: Nonterminal, prod_seq: List[Production]) -> None:
            plist = self._nt_dict[nt]
            for _, p in plist:
                for s in p:
                    if isinstance(s, Terminal):
                        follow[s].append(prod_seq + [p])
                        break
                    if s not in seen:
                        assert isinstance(s, Nonterminal)
                        seen.add(s)
                        add_follow(s, prod_seq + [p])
                    if s not in nullable:
                        break

        seen.add(nonterminal)
        add_follow(nonterminal, [])

        for terminal, pseq in follow.items():
            print("Terminal {0} can follow by these productions:".format(terminal))
            for p in pseq:
                print("   {0}".format(str(p)))
            print()

        return follow


if __name__ == "__main__":

    # If run as a main program, do a verbose grammar check

    import sys

    try:
        # Read configuration file
        basepath, _ = os.path.split(os.path.realpath(__file__))
        Settings.read(os.path.join(basepath, "config", "GreynirPackage.conf"))
    except ConfigError as e:
        print("Configuration error: {0}".format(e))
        quit()

    from tokenizer.abbrev import Abbreviations

    # Make sure that the tokenizer has loaded its abbreviations
    # before we parse the grammar, since they are used in error checking
    Abbreviations.initialize()

    fname = "Greynir.grammar"
    args = sys.argv
    if len(args) == 2:
        fname = args[1]

    ts = os.path.getmtime(fname)
    if ts is None:
        print("Unable to read grammar file {0}".format(fname))
    else:
        print(
            "Reading grammar file {0} with timestamp {1:%Y-%m-%d %H:%M:%S}\n".format(
                fname, datetime.fromtimestamp(ts)
            )
        )
        import time

        t0 = time.time()
        g = Grammar()
        try:
            g.read(fname, verbose=True)
            print(
                "Grammar parsed and loaded in {0:.2f} seconds".format(time.time() - t0)
            )
            print(
                "Grammar has {0:,} terminals, {1:,} nonterminals and {2:,} productions".format(
                    g.num_terminals, g.num_nonterminals, g.num_productions
                )
            )
        except GrammarError as err:
            print(str(err))
